from rdflib import Graph

from dsp_tools.commands.validate_data.sparql.value_shacl import construct_property_shapes


def construct_content_shapes_graph(onto: Graph) -> Graph:
    """
    Constructs a shapes graph from a project ontology

    Args:
        onto: ontology as graph

    Returns:
        shapes graph
    """
    return construct_property_shapes(onto)



from dataclasses import dataclass
from typing import Optional
from typing import Union

from dsp_tools.models.exceptions import BaseError

# ruff: noqa: E501 (line-too-long)


@dataclass(frozen=True)
class PropertyElement:
    """
    A PropertyElement object carries more information about a property value than the value itself.
    The "value" is the value that could be passed to a method as plain string/int/float/bool. Use a PropertyElement
    instead to define more precisely what attributes your value tag (e.g. `<text>`, `<uri>`, ...) will have.

    Args:
        value: This is the content that will be written into the value tag (e.g. `<text>`, `<uri>`, ...)
        permissions: This is the permissions that your `<text>` tag (for example) will have
        comment: This is the comment that your `<text>` tag (for example) will have
        encoding: For `<text>` tags only. If provided, it must be "xml" or "utf8".

    Examples:
        See the difference between the first and the second example:

        >>> excel2xml.make_text_prop(":testproperty", "first text")
                <text-prop name=":testproperty">
                    <text encoding="utf8" permissions="prop-default">
                        first text
                    </text>
                </text-prop>
        >>> excel2xml.make_text_prop(":testproperty", excel2xml.PropertyElement("first text", permissions="prop-restricted", encoding="xml"))
                <text-prop name=":testproperty">
                    <text encoding="xml" permissions="prop-restricted">
                        first text
                    </text>
                </text-prop>
    """

    value: Union[str, int, float, bool]
    permissions: str = "prop-default"
    comment: Optional[str] = None
    encoding: Optional[str] = None

    def __post_init__(self) -> None:
        if self.encoding not in ["utf8", "xml", None]:
            raise BaseError(f"'{self.encoding}' is not a valid encoding for a PropertyElement")



import pytest
from lxml import etree

from dsp_tools.commands.validate_data.deserialise_input import deserialise_xml
from dsp_tools.commands.validate_data.models.data_deserialised import ProjectDeserialised


def test_to_data_rdf(data_xml: etree._Element) -> None:
    res = deserialise_xml(data_xml)
    assert isinstance(res, ProjectDeserialised)
    assert res.info.shortcode == "9999"
    assert res.info.default_onto == "onto"
    assert len(res.data.resources) == 14


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import json
import unicodedata
from pathlib import Path
from typing import Any
from typing import Optional
from typing import TypeGuard
from typing import Union

import pandas as pd
import regex

from dsp_tools.commands.excel2xml.propertyelement import PropertyElement
from dsp_tools.models.exceptions import BaseError


def simplify_name(value: str) -> str:
    """
    Simplifies a given value in order to use it as node name

    Args:
        value: The value to be simplified

    Returns:
        str: The simplified value
    """
    simplified_value = value.lower()

    # normalize characters (p.ex. ä becomes a)
    simplified_value = unicodedata.normalize("NFKD", simplified_value)

    # replace forward slash and whitespace with a dash
    simplified_value = regex.sub("[/\\s]+", "-", simplified_value)

    # delete all characters which are not letters, numbers or dashes
    simplified_value = regex.sub("[^A-Za-z0-9\\-]+", "", simplified_value)

    return simplified_value


def check_notna(value: Optional[Any]) -> TypeGuard[Any]:
    """
    Check a value if it is usable in the context of data archiving. A value is considered usable if it is
     - a number (integer or float, but not np.nan)
     - a boolean
     - a string with at least one Unicode letter (matching the regex ``\\p{L}``) or number, or at least one _, !, or ?
       (The strings `None`, `<NA>`, `N/A`, and `-` are considered invalid.)
     - a PropertyElement whose "value" fulfills the above criteria

    Args:
        value: any object encountered when analysing data

    Returns:
        True if the value is usable, False if it is N/A or otherwise unusable

    Examples:
        >>> check_notna(0)      == True
        >>> check_notna(False)  == True
        >>> check_notna("œ")    == True
        >>> check_notna("0")    == True
        >>> check_notna("_")    == True
        >>> check_notna("!")    == True
        >>> check_notna("?")    == True
        >>> check_notna(None)   == False
        >>> check_notna("None") == False
        >>> check_notna(<NA>)   == False
        >>> check_notna("<NA>") == False
        >>> check_notna("-")    == False
        >>> check_notna(" ")    == False
    """

    if isinstance(value, PropertyElement):
        value = value.value

    if isinstance(value, (bool, int)) or (
        isinstance(value, float) and pd.notna(value)
    ):  # necessary because isinstance(np.nan, float)
        return True
    elif isinstance(value, str):
        return bool(regex.search(r"[\p{L}\d_!?]", value, flags=regex.UNICODE)) and not bool(
            regex.search(r"^(none|<NA>|-|n/a)$", value, flags=regex.IGNORECASE)
        )
    else:
        return False


def parse_json_input(project_file_as_path_or_parsed: Union[str, Path, dict[str, Any]]) -> dict[str, Any]:
    """
    Check the input for a method that expects a JSON project definition, either as file path or as parsed JSON object:
    If it is parsed already, return it unchanged.
    If the input is a file path, parse it.

    Args:
        project_file_as_path_or_parsed: path to the JSON project definition, or parsed JSON object

    Raises:
        BaseError: if the input is invalid

    Returns:
        the parsed JSON object
    """
    project_definition: dict[str, Any] = {}
    if isinstance(project_file_as_path_or_parsed, dict):
        project_definition = project_file_as_path_or_parsed
    elif isinstance(project_file_as_path_or_parsed, (str, Path)) and Path(project_file_as_path_or_parsed).exists():
        with open(project_file_as_path_or_parsed, encoding="utf-8") as f:
            try:
                project_definition = json.load(f)
            except json.JSONDecodeError as e:
                msg = f"The input file '{project_file_as_path_or_parsed}' cannot be parsed to a JSON object."
                raise BaseError(msg) from e
    else:
        raise BaseError("Invalid input: The input must be a path to a JSON file or a parsed JSON object.")
    return project_definition



import regex
from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib.term import Node

from dsp_tools.commands.validate_data.models.input_problems import AllProblems
from dsp_tools.commands.validate_data.models.input_problems import ContentRegexViolation
from dsp_tools.commands.validate_data.models.input_problems import InputProblem
from dsp_tools.commands.validate_data.models.input_problems import MaxCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import MinCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import NonExistentCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import UnexpectedResults
from dsp_tools.commands.validate_data.models.input_problems import ValueTypeViolation
from dsp_tools.commands.validate_data.models.validation import CardinalityValidationResult
from dsp_tools.commands.validate_data.models.validation import ContentValidationResult
from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.commands.validate_data.models.validation import ValidationReports


def reformat_validation_graph(report: ValidationReports) -> AllProblems:
    """
    Reformats the validation result from an RDF graph into class instances
    that are used to communicate the problems with the user.

    Args:
        report: with all the information necessary to construct a user message

    Returns:
        All Problems
    """
    reformatted_results: list[InputProblem] = []
    unexpected_components: list[UnexpectedComponent] = []
    if report.cardinality_validation:
        reformatted, unexpected = _get_cardinality_input_errors(report.cardinality_validation, report.data_graph)
        reformatted_results.extend(reformatted)
        unexpected_components.extend(unexpected)
    if report.content_validation:
        reformatted, unexpected = _get_content_input_errors(report.content_validation, report.data_graph)
        reformatted_results.extend(reformatted)
        unexpected_components.extend(unexpected)
    unexpected_found = UnexpectedResults(unexpected_components) if unexpected_components else None
    return AllProblems(reformatted_results, unexpected_found)


def _get_cardinality_input_errors(
    results_graph: Graph, data_graph: Graph
) -> tuple[list[InputProblem], list[UnexpectedComponent]]:
    validation_results = _query_for_cardinality_validation_results(results_graph, data_graph)
    input_problems: list[InputProblem] = []
    unexpected_components: list[UnexpectedComponent] = []
    for violation in validation_results:
        problem = _reformat_one_cardinality_validation_result(violation)
        if isinstance(problem, UnexpectedComponent):
            unexpected_components.append(problem)
        else:
            input_problems.append(problem)
    return input_problems, unexpected_components


def _query_for_cardinality_validation_results(
    results_graph: Graph, data_graph: Graph
) -> list[CardinalityValidationResult]:
    violations = results_graph.subjects(RDF.type, SH.ValidationResult)
    return [_query_for_one_cardinality_validation_result(x, results_graph, data_graph) for x in violations]


def _query_for_one_cardinality_validation_result(
    bn: Node, results_graph: Graph, data_graph: Graph
) -> CardinalityValidationResult:
    focus_nd = next(results_graph.objects(bn, SH.focusNode))
    res_type = next(data_graph.objects(focus_nd, RDF.type))
    path = next(results_graph.objects(bn, SH.resultPath))
    component = next(results_graph.objects(bn, SH.sourceConstraintComponent))
    msg = str(next(results_graph.objects(bn, SH.resultMessage)))
    return CardinalityValidationResult(
        source_constraint_component=component,
        res_iri=focus_nd,
        res_class=res_type,
        property=path,
        results_message=msg,
    )


def _reformat_one_cardinality_validation_result(
    violation: CardinalityValidationResult,
) -> InputProblem | UnexpectedComponent:
    subject_id = _reformat_data_iri(str(violation.res_iri))
    prop_name = _reformat_onto_iri(str(violation.property))
    res_type = _reformat_onto_iri(str(violation.res_class))
    match violation.source_constraint_component:
        case SH.MaxCountConstraintComponent:
            return MaxCardinalityViolation(
                res_id=subject_id,
                res_type=res_type,
                prop_name=prop_name,
                expected_cardinality=violation.results_message,
            )
        case SH.MinCountConstraintComponent:
            return MinCardinalityViolation(
                res_id=subject_id,
                res_type=res_type,
                prop_name=prop_name,
                expected_cardinality=violation.results_message,
            )
        case SH.ClosedConstraintComponent:
            return NonExistentCardinalityViolation(
                res_id=subject_id,
                res_type=res_type,
                prop_name=prop_name,
            )
        case _:
            return UnexpectedComponent(str(violation.source_constraint_component))


def _get_content_input_errors(
    results_graph: Graph, data_graph: Graph
) -> tuple[list[InputProblem], list[UnexpectedComponent]]:
    violations = _query_for_content_validation_results(results_graph, data_graph)

    input_problems: list[InputProblem] = []
    unexpected_components: list[UnexpectedComponent] = []
    for violation in violations:
        problem = _reformat_one_content_validation_result(violation)
        if isinstance(problem, UnexpectedComponent):
            unexpected_components.append(problem)
        else:
            input_problems.append(problem)
    return input_problems, unexpected_components


def _query_for_content_validation_results(results_graph: Graph, data_graph: Graph) -> list[ContentValidationResult]:
    main_results = list(results_graph.subjects(SH.sourceConstraintComponent, SH.NodeConstraintComponent))
    return [_query_for_one_content_validation_result(x, results_graph, data_graph) for x in main_results]


def _query_for_one_content_validation_result(
    bn: Node, results_graph: Graph, data_graph: Graph
) -> ContentValidationResult:
    focus_nd = next(results_graph.objects(bn, SH.focusNode))
    res_type = next(data_graph.objects(focus_nd, RDF.type))
    path = next(results_graph.objects(bn, SH.resultPath))
    value_iri = next(results_graph.objects(bn, SH.value))
    value_type = next(data_graph.objects(value_iri, RDF.type))
    component = next(results_graph.objects(bn, SH.sourceConstraintComponent))
    detail_bn = next(results_graph.objects(bn, SH.detail))
    detail_component = next(results_graph.objects(detail_bn, SH.sourceConstraintComponent))
    val = None
    if node_value := list(results_graph.objects(detail_bn, SH.value)):
        val = str(node_value[0])
    msg = str(next(results_graph.objects(detail_bn, SH.resultMessage)))
    return ContentValidationResult(
        source_constraint_component=component,
        res_iri=focus_nd,
        res_class=res_type,
        property=path,
        results_message=msg,
        detail_bn_component=detail_component,
        value_type=value_type,
        value=val,
    )


def _reformat_one_content_validation_result(val_result: ContentValidationResult) -> InputProblem | UnexpectedComponent:
    subject_id = _reformat_data_iri(str(val_result.res_iri))
    prop_name = _reformat_onto_iri(str(val_result.property))
    res_type = _reformat_onto_iri(str(val_result.res_class))
    match val_result.detail_bn_component:
        case SH.PatternConstraintComponent:
            val: str | None = val_result.value
            if val and not regex.search(r"\S+", val):
                val = None
            return ContentRegexViolation(
                res_id=subject_id,
                res_type=res_type,
                prop_name=prop_name,
                expected_format=val_result.results_message,
                actual_content=val,
            )
        case SH.ClassConstraintComponent | SH.MinCountConstraintComponent:
            actual_type = _reformat_onto_iri(str(val_result.value_type)).replace("knora-api:", "")
            return ValueTypeViolation(
                res_id=subject_id,
                res_type=res_type,
                prop_name=prop_name,
                actual_type=actual_type,
                expected_type=val_result.results_message,
            )
        case _:
            return UnexpectedComponent(str(val_result.source_constraint_component))


def _reformat_onto_iri(prop: str) -> str:
    onto = prop.split("/")[-2]
    return f'{onto}:{prop.split("#")[-1]}'


def _reformat_data_iri(iri: str) -> str:
    return iri.replace("http://data/", "")



from test.unittests.commands.validate_data.fixtures_data_deserialised import *  # noqa: F403
from test.unittests.commands.validate_data.fixtures_data_rdf import *  # noqa: F403
from test.unittests.commands.validate_data.fixtures_onto import *  # noqa: F403
from test.unittests.commands.validate_data.fixtures_validation_result import *  # noqa: F403
from test.unittests.commands.validate_data.fixtures_xml_data import *  # noqa: F403



from __future__ import annotations

import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Protocol

from lxml import etree

from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.uri_util import is_iiif_uri
from dsp_tools.xmllib.models.user_enums import Permissions
from dsp_tools.xmllib.value_checkers import is_string_like

XML_NAMESPACE_MAP = {None: "https://dasch.swiss/schema", "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
DASCH_SCHEMA = "{https://dasch.swiss/schema}"


class AbstractFileValue(Protocol):
    value: str | Path
    permissions: Permissions
    comment: str | None = None

    def serialise(self) -> etree._Element:
        raise NotImplementedError


@dataclass
class FileValue(AbstractFileValue):
    value: str | Path
    permissions: Permissions = Permissions.DOAP
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_string_like(str(self.value)):
            _warn_type_mismatch(expected_type="file name", value=self.value, res_id=self.resource_id)

    def serialise(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.DOAP:
            attribs["permissions"] = self.permissions.value
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}bitstream", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


@dataclass
class IIIFUri(AbstractFileValue):
    value: str
    permissions: Permissions = Permissions.DOAP
    comment: str | None = None
    resource_id: str | None = None

    def __post_init__(self) -> None:
        if not is_iiif_uri(self.value):
            _warn_type_mismatch(expected_type="IIIF uri", value=self.value, res_id=self.resource_id)

    def serialise(self) -> etree._Element:
        attribs = {}
        if self.permissions != Permissions.DOAP:
            attribs["permissions"] = self.permissions.value
        if self.comment:
            attribs["comment"] = self.comment
        ele = etree.Element(f"{DASCH_SCHEMA}iiif-uri", attrib=attribs, nsmap=XML_NAMESPACE_MAP)
        ele.text = str(self.value)
        return ele


def _warn_type_mismatch(expected_type: str, value: Any, res_id: str | None) -> None:
    """Emits a warning if a values is not in the expected format."""
    if res_id:
        msg = f"The Resource '{res_id}' has an invalid input: The value '{value}' is not a valid {expected_type}."
    else:
        msg = f"The value '{value}' is not a valid {expected_type}."
    warnings.warn(DspToolsUserWarning(msg))



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass


@dataclass
class ProjectDeserialised:
    info: ProjectInformation
    data: DataDeserialised


@dataclass
class ProjectInformation:
    shortcode: str
    default_onto: str


@dataclass
class DataDeserialised:
    resources: list[AbstractResource]


@dataclass
class AbstractResource(ABC):
    res_id: str
    label: str


@dataclass
class ResourceDeserialised(AbstractResource):
    res_class: str
    values: list[ValueDeserialised]


@dataclass
class AnnotationDeserialised(AbstractResource): ...


@dataclass
class RegionDeserialised(AbstractResource): ...


@dataclass
class LinkObjDeserialised(AbstractResource): ...


@dataclass
class VideoSegmentDeserialised(AbstractResource): ...


@dataclass
class AudioSegmentDeserialised(AbstractResource): ...


@dataclass
class ValueDeserialised:
    prop_name: str
    object_value: str | None


@dataclass
class BooleanValueDeserialised(ValueDeserialised): ...


@dataclass
class ColorValueDeserialised(ValueDeserialised): ...


@dataclass
class DateValueDeserialised(ValueDeserialised): ...


@dataclass
class DecimalValueDeserialised(ValueDeserialised): ...


@dataclass
class GeonameValueDeserialised(ValueDeserialised): ...


@dataclass
class IntValueDeserialised(ValueDeserialised): ...


@dataclass
class LinkValueDeserialised(ValueDeserialised): ...


@dataclass
class ListValueDeserialised(ValueDeserialised):
    list_name: str


@dataclass
class SimpleTextDeserialised(ValueDeserialised): ...


@dataclass
class RichtextDeserialised(ValueDeserialised): ...


@dataclass
class TimeValueDeserialised(ValueDeserialised): ...


@dataclass
class UriValueDeserialised(ValueDeserialised): ...



from __future__ import annotations

import importlib.resources
import json
from pathlib import Path
from typing import Any
from typing import Union

import jsonpath_ng
import jsonpath_ng.ext
import jsonschema
import networkx as nx
import regex

from dsp_tools.commands.excel2json.lists import expand_lists_from_excel
from dsp_tools.models.exceptions import BaseError


def _check_for_duplicate_names(project_definition: dict[str, Any]) -> bool:
    """
    Check that the resource names and property names are unique.

    Args:
        project_definition: parsed JSON project definition

    Raises:
        BaseError: detailed error message if there is a duplicate resource name / property name

    Returns:
        True if the resource/property names are unique
    """
    propnames_duplicates, resnames_duplicates = _find_duplicates(project_definition)

    if not resnames_duplicates and not propnames_duplicates:
        return True

    err_msg = "Resource names and property names must be unique inside every ontology.\n"
    for ontoname, res_duplicates in resnames_duplicates.items():
        for res_duplicate in sorted(res_duplicates):
            err_msg += f"Resource '{res_duplicate}' appears multiple times in the ontology '{ontoname}'.\n"
    for ontoname, prop_duplicates in propnames_duplicates.items():
        for prop_duplicate in sorted(prop_duplicates):
            err_msg += f"Property '{prop_duplicate}' appears multiple times in the ontology '{ontoname}'.\n"

    raise BaseError(err_msg)


def _find_duplicates(project_definition: dict[str, Any]) -> tuple[dict[str, set[str]], dict[str, set[str]]]:
    resnames_duplicates: dict[str, set[str]] = {}
    propnames_duplicates: dict[str, set[str]] = {}
    for onto in project_definition["project"]["ontologies"]:
        resnames = [r["name"] for r in onto["resources"]]
        if len(set(resnames)) != len(resnames):
            for elem in resnames:
                if resnames.count(elem) > 1:
                    if resnames_duplicates.get(onto["name"]):
                        resnames_duplicates[onto["name"]].add(elem)
                    else:
                        resnames_duplicates[onto["name"]] = {elem}

        propnames = [p["name"] for p in onto["properties"]]
        if len(set(propnames)) != len(propnames):
            for elem in propnames:
                if propnames.count(elem) > 1:
                    if propnames_duplicates.get(onto["name"]):
                        propnames_duplicates[onto["name"]].add(elem)
                    else:
                        propnames_duplicates[onto["name"]] = {elem}
    return propnames_duplicates, resnames_duplicates


def _check_for_undefined_super_resource(project_definition: dict[str, Any]) -> bool:
    """
    Check the superresources that claim to point to a resource defined in the same JSON project.
    Check if the resource they point to actually exists.
    (DSP base resources and resources from other ontologies are not considered.)

    Args:
        project_definition: parsed JSON project definition

    Raises:
        BaseError: detailed error message if a superresource is not existent

    Returns:
        True if the superresource are valid
    """
    errors: dict[str, list[str]] = {}
    for onto in project_definition["project"]["ontologies"]:
        ontoname = onto["name"]
        resnames = [r["name"] for r in onto["resources"]]
        for res in onto["resources"]:
            supers = res["super"] if isinstance(res["super"], list) else [res["super"]]
            # form of supers:
            #  - Resource      # DSP base resource
            #  - other:res     # other onto
            #  - same:res      # same onto
            #  - :res          # same onto (short form)

            # filter out DSP base resources
            supers = [s for s in supers if ":" in s]
            # extend short form
            supers = [regex.sub(r"^:", f"{ontoname}:", s) for s in supers]
            # filter out other ontos
            supers = [s for s in supers if regex.search(f"^{ontoname}:", s)]
            # convert to short form
            supers = [regex.sub(f"^{ontoname}", "", s) for s in supers]

            if invalid_references := [s for s in supers if regex.sub(":", "", s) not in resnames]:
                errors[f"Ontology '{ontoname}', resource '{res['name']}'"] = invalid_references

    if errors:
        err_msg = "Your data model contains resources that are derived from an invalid super-resource:\n" + "\n".join(
            f" - {loc}: {invalids}" for loc, invalids in errors.items()
        )
        raise BaseError(err_msg)
    return True


def _check_for_undefined_super_property(project_definition: dict[str, Any]) -> bool:
    """
    Check the superproperties that claim to point to a property defined in the same JSON project.
    Check if the property they point to actually exists.
    (DSP base properties and properties from other ontologies are not considered.)

    Args:
        project_definition: parsed JSON project definition

    Raises:
        BaseError: detailed error message if a superproperty is not existent

    Returns:
        True if the superproperties are valid
    """
    errors: dict[str, list[str]] = {}
    for onto in project_definition["project"]["ontologies"]:
        ontoname = onto["name"]
        propnames = [p["name"] for p in onto["properties"]]
        for prop in onto["properties"]:
            supers = prop["super"]
            # form of supers:
            #  - isSegmentOf   # DSP base property
            #  - other:prop    # other onto
            #  - same:prop     # same onto
            #  - :prop         # same onto (short form)

            # filter out DSP base properties
            supers = [s for s in supers if ":" in s]
            # extend short form
            supers = [regex.sub(r"^:", f"{ontoname}:", s) for s in supers]
            # filter out other ontos
            supers = [s for s in supers if regex.search(f"^{ontoname}:", s)]
            # convert to short form
            supers = [regex.sub(f"^{ontoname}", "", s) for s in supers]

            if invalid_references := [s for s in supers if regex.sub(":", "", s) not in propnames]:
                errors[f"Ontology '{ontoname}', property '{prop['name']}'"] = invalid_references

    if errors:
        err_msg = "Your data model contains properties that are derived from an invalid super-property:\n" + "\n".join(
            f" - {loc}: {invalids}" for loc, invalids in errors.items()
        )
        raise BaseError(err_msg)
    return True


def _check_for_undefined_cardinalities(project_definition: dict[str, Any]) -> bool:
    """
    Check if the propnames that are used in the cardinalities of each resource are defined in the "properties"
    section. (DSP base properties and properties from other ontologies are not considered.)

    Args:
        project_definition: parsed JSON project definition

    Raises:
        BaseError: detailed error message if a cardinality is used that is not defined

    Returns:
        True if all cardinalities are defined in the "properties" section
    """
    errors: dict[str, list[str]] = {}
    for onto in project_definition["project"]["ontologies"]:
        ontoname = onto["name"]
        propnames = [prop["name"] for prop in onto["properties"]]
        for res in onto["resources"]:
            cardnames = [card["propname"] for card in res.get("cardinalities", [])]
            # form of the cardnames:
            #  - isSegmentOf   # DSP base property
            #  - other:prop    # other onto
            #  - same:prop     # same onto
            #  - :prop         # same onto (short form)

            # filter out DSP base properties
            cardnames = [card for card in cardnames if ":" in card]
            # extend short form
            cardnames = [regex.sub(r"^:", f"{ontoname}:", card) for card in cardnames]
            # filter out other ontos
            cardnames = [card for card in cardnames if regex.search(f"^{ontoname}:", card)]
            # convert to short form
            cardnames = [regex.sub(f"^{ontoname}:", ":", card) for card in cardnames]

            if invalid_cardnames := [card for card in cardnames if regex.sub(":", "", card) not in propnames]:
                errors[f"Ontology '{ontoname}', resource '{res['name']}'"] = invalid_cardnames

    if errors:
        err_msg = "Your data model contains cardinalities with invalid propnames:\n" + "\n".join(
            f" - {loc}: {invalids}" for loc, invalids in errors.items()
        )
        raise BaseError(err_msg)
    return True


def _check_for_deprecated_syntax(project_definition: dict[str, Any]) -> bool:  # noqa: ARG001 (unused argument)
    return True


def validate_project(
    input_file_or_json: Union[dict[str, Any], str],
    expand_lists: bool = True,
) -> bool:
    """
    Validates a JSON project definition file.

    First, the Excel file references in the "lists" section are expanded
    (unless this behaviour is disabled).

    Then, the project is validated against the JSON schema.

    Next, some checks are performed that are too complex for JSON schema.

    At last, a check is performed
    if this project's ontologies contain properties derived from hasLinkTo
    that form a circular reference.
    If so, these properties must have the cardinality 0-1 or 0-n,
    because during the xmlupload process,
    these values are temporarily removed.

    Args:
        input_file_or_json: the project to be validated, can either be a file path or a parsed JSON file
        expand_lists: if True, the Excel file references in the "lists" section will be expanded

    Raises:
        BaseError: detailed error report if the validation doesn't pass

    Returns:
        True if the project passed validation.
    """

    # parse input
    if isinstance(input_file_or_json, dict) and "project" in input_file_or_json:
        project_definition = input_file_or_json
    elif (
        isinstance(input_file_or_json, str)
        and Path(input_file_or_json).is_file()
        and regex.search(r"\.json$", input_file_or_json)
    ):
        with open(input_file_or_json, encoding="utf-8") as f:
            project_definition = json.load(f)
    else:
        raise BaseError(f"Input '{input_file_or_json}' is neither a file path nor a JSON object.")

    # expand all lists referenced in the "lists" section of the project definition,
    # and add them to the project definition
    if expand_lists:
        if new_lists := expand_lists_from_excel(project_definition["project"].get("lists", [])):
            project_definition["project"]["lists"] = new_lists

    # validate the project definition against the schema
    with (
        importlib.resources.files("dsp_tools")
        .joinpath("resources/schema/project.json")
        .open(encoding="utf-8") as schema_file
    ):
        project_schema = json.load(schema_file)
    try:
        jsonschema.validate(instance=project_definition, schema=project_schema)
    except jsonschema.ValidationError as err:
        raise BaseError(
            f"The JSON project file cannot be created due to the following validation error: {err.message}.\n"
            f"The error occurred at {err.json_path}:\n{err.instance}"
        ) from None

    # make some checks that are too complex for JSON schema
    _check_for_undefined_super_property(project_definition)
    _check_for_undefined_super_resource(project_definition)
    _check_for_undefined_cardinalities(project_definition)
    _check_for_duplicate_names(project_definition)
    _check_for_deprecated_syntax(project_definition)

    # cardinalities check for circular references
    return _check_cardinalities_of_circular_references(project_definition)


def _check_cardinalities_of_circular_references(project_definition: dict[Any, Any]) -> bool:
    """
    Check a JSON project file if it contains properties derived from hasLinkTo that form a circular reference. If so,
    these properties must have the cardinality 0-1 or 0-n, because during the xmlupload process, these values
    are temporarily removed.

    Args:
        project_definition: dictionary with a DSP project (as defined in a JSON project file)

    Raises:
        BaseError: if there is a circle with at least one element that has a cardinality of "1" or "1-n"

    Returns:
        True if no circle was detected, or if all elements of all circles are of cardinality "0-1" or "0-n".
    """

    link_properties = _collect_link_properties(project_definition)
    errors = _identify_problematic_cardinalities(project_definition, link_properties)

    if len(errors) == 0:
        return True

    error_message = (
        "ERROR: Your ontology contains properties derived from 'hasLinkTo' that allow circular references "
        "between resources. This is not a problem in itself, but if you try to upload data that actually "
        "contains circular references, these 'hasLinkTo' properties will be temporarily removed from the "
        "affected resources. Therefore, it is necessary that all involved 'hasLinkTo' properties have a "
        "cardinality of 0-1 or 0-n. \n"
        "Please make sure that the following properties have a cardinality of 0-1 or 0-n:"
    )
    for error in errors:
        error_message = f"{error_message}\n    - Resource {error[0]}, property {error[1]}"
    raise BaseError(error_message)


def _collect_link_properties(project_definition: dict[Any, Any]) -> dict[str, list[str]]:
    """
    Maps the properties derived from hasLinkTo to the resource classes they point to.

    Args:
        project_definition: parsed JSON file

    Returns:
        A (possibly empty) dictionary in the form {"rosetta:hasImage2D": ["rosetta:Image2D"], ...}
    """
    ontos = project_definition["project"]["ontologies"]
    hasLinkTo_props = {"hasLinkTo", "isPartOf", "isRegionOf", "isAnnotationOf"}
    link_properties: dict[str, list[str]] = {}
    for index, onto in enumerate(ontos):
        hasLinkTo_matches = []
        # look for child-properties down to 5 inheritance levels that are derived from hasLinkTo-properties
        for _ in range(5):
            for hasLinkTo_prop in hasLinkTo_props:
                hasLinkTo_matches.extend(
                    jsonpath_ng.ext.parse(
                        f"$.project.ontologies[{index}].properties[?super[*] == {hasLinkTo_prop}]"
                    ).find(project_definition)
                )
            # make the children from this iteration to the parents of the next iteration
            hasLinkTo_props = {x.value["name"] for x in hasLinkTo_matches}
        prop_obj_pair: dict[str, list[str]] = {}
        for match in hasLinkTo_matches:
            prop = onto["name"] + ":" + match.value["name"]
            target = match.value["object"]
            if target != "Resource":
                # make the target a fully qualified name (with the ontology's name prefixed)
                target = regex.sub(r"^:([^:]+)$", f"{onto['name']}:\\1", target)
            prop_obj_pair[prop] = [target]
        link_properties.update(prop_obj_pair)

    # in case the object of a property is "Resource", the link can point to any resource class
    all_res_names: list[str] = []
    for onto in ontos:
        matches = jsonpath_ng.ext.parse("$.resources[*].name").find(onto)
        tmp = [f"{onto['name']}:{match.value}" for match in matches]
        all_res_names.extend(tmp)
    for prop, targ in link_properties.items():
        if "Resource" in targ:
            link_properties[prop] = all_res_names

    return link_properties


def _identify_problematic_cardinalities(
    project_definition: dict[Any, Any],
    link_properties: dict[str, list[str]],
) -> list[tuple[str, str]]:
    """
    Make an error list with all cardinalities that are part of a circle but have a cardinality of "1" or "1-n".

    Args:
        project_definition: parsed JSON file
        link_properties: mapping of hasLinkTo-properties to classes they point to,
            e.g. {"rosetta:hasImage2D": ["rosetta:Image2D"], ...}

    Returns:
        a (possibly empty) list of (resource, problematic_cardinality) tuples
    """
    cardinalities, dependencies = _extract_cardinalities_from_project(project_definition, link_properties)
    graph = _make_cardinality_dependency_graph(dependencies)
    errors = _find_circles_with_min_one_cardinality(graph, cardinalities, dependencies)
    return sorted(errors, key=lambda x: x[0])


def _extract_cardinalities_from_project(
    project_definition: dict[Any, Any],
    link_properties: dict[str, list[str]],
) -> tuple[dict[str, dict[str, str]], dict[str, dict[str, list[str]]]]:
    # dependencies = {"rosetta:Text": {"rosetta:hasImage2D": ["rosetta:Image2D"], ...}}
    dependencies: dict[str, dict[str, list[str]]] = {}
    # cardinalities = {"rosetta:Text": {"rosetta:hasImage2D": "0-1", ...}}
    cardinalities: dict[str, dict[str, str]] = {}

    for onto in project_definition["project"]["ontologies"]:
        for resource in onto["resources"]:
            resname: str = onto["name"] + ":" + resource["name"]
            for card in resource.get("cardinalities", []):
                # make the cardinality a fully qualified name (with the ontology's name prefixed)
                cardname = regex.sub(r"^(:?)([^:]+)$", f"{onto['name']}:\\2", card["propname"])
                if cardname in link_properties:
                    # Look out: if `targets` is created with `targets = link_properties[cardname]`, the ex-
                    # pression `dependencies[resname][cardname] = targets` causes `dependencies[resname][cardname]`
                    # to point to `link_properties[cardname]`. Due to that, the expression
                    # `dependencies[resname][cardname].extend(targets)` will modify "link_properties"!
                    # For this reason, `targets` must be created with `targets = list(link_properties[cardname])`
                    targets = list(link_properties[cardname])
                    if resname not in dependencies:
                        dependencies[resname] = {cardname: targets}
                        cardinalities[resname] = {cardname: card["cardinality"]}
                    elif cardname not in dependencies[resname]:
                        dependencies[resname][cardname] = targets
                        cardinalities[resname][cardname] = card["cardinality"]
                    else:
                        dependencies[resname][cardname].extend(targets)
    return cardinalities, dependencies


def _make_cardinality_dependency_graph(dependencies: dict[str, dict[str, list[str]]]) -> nx.MultiDiGraph[Any]:
    graph: nx.MultiDiGraph[Any] = nx.MultiDiGraph()
    for start, cards in dependencies.items():
        for edge, targets in cards.items():
            for target in targets:
                graph.add_edge(start, target, edge)
    return graph


def _find_circles_with_min_one_cardinality(
    graph: nx.MultiDiGraph[Any], cardinalities: dict[str, dict[str, str]], dependencies: dict[str, dict[str, list[str]]]
) -> set[tuple[str, str]]:
    errors: set[tuple[str, str]] = set()
    circles = list(nx.algorithms.cycles.simple_cycles(graph))
    for circle in circles:
        for index, resource in enumerate(circle):
            target = circle[(index + 1) % len(circle)]
            prop = ""
            for _property, targets in dependencies[resource].items():
                if target in targets:
                    prop = _property
            if cardinalities[resource].get(prop) not in ["0-1", "0-n"]:
                errors.add((resource, prop))
    return errors



from rdflib import Graph


def construct_property_shapes(onto: Graph) -> Graph:
    """
    Returns the sh:PropertyShape for the properties in the ontology.

    Args:
        onto: ontology graph

    Returns:
        Graph with the property shapes
    """
    g = Graph()
    g += _construct_property_type_shape_based_on_object_type(onto)
    g += _construct_link_value_shape(onto)
    g += _construct_property_type_text_value(onto)
    return g + _add_property_shapes_to_class_shapes(onto)


def _add_property_shapes_to_class_shapes(onto: Graph) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#> 
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>

    CONSTRUCT {

      ?classShapeIRI sh:property ?propShapesIRI ;
                     sh:targetClass ?class .

    } WHERE {

      ?class a owl:Class ;
          knora-api:isResourceClass true ;
          knora-api:canBeInstantiated true ;
          rdfs:subClassOf ?restriction .
      ?restriction a owl:Restriction ;
          owl:onProperty ?propRestriction ;
          salsah-gui:guiOrder ?order .
      FILTER NOT EXISTS { ?propRestriction knora-api:isLinkValueProperty true }

      BIND(IRI(CONCAT(str(?class), "_Shape")) AS ?classShapeIRI)
      BIND(IRI(CONCAT(str(?propRestriction), "_PropShape")) AS ?propShapesIRI)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_property_type_shape_based_on_object_type(onto: Graph) -> Graph:
    def as_object_type_and_shacl_shape(property_type: str) -> tuple[str, str]:
        return "knora-api:" + property_type + "Value", "api-shapes:" + property_type + "Value_ClassShape"

    property_types = {"Boolean", "Color", "Date", "Decimal", "Geoname", "Int", "List", "Time", "Uri"}

    g = Graph()
    for t in property_types:
        object_type, shacl_shape = as_object_type_and_shacl_shape(t)
        g += _construct_one_property_type_shape_based_on_object_type(onto, object_type, shacl_shape)
    return g


def _construct_one_property_type_shape_based_on_object_type(onto: Graph, object_type: str, shacl_shape: str) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>
    
    CONSTRUCT {
        
        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node %(shacl_shape)s .
    
    } WHERE {
      
        ?prop a owl:ObjectProperty ;
                knora-api:objectType %(object_type)s .
      
        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
    }
    """ % {"object_type": object_type, "shacl_shape": shacl_shape}  # noqa: UP031 (printf-string-formatting)
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_link_value_shape(onto: Graph) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {

        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node api-shapes:LinkValue_ClassShape .

    } WHERE {

        ?prop a owl:ObjectProperty ;
                salsah-gui:guiElement salsah-gui:Searchbox .

        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
    }
    """
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()


def _construct_property_type_text_value(onto: Graph) -> Graph:
    property_type_mapper = {
        "salsah-gui:SimpleText": "api-shapes:SimpleTextValue_ClassShape",
        "salsah-gui:Textarea": "api-shapes:SimpleTextValue_ClassShape",
        "salsah-gui:Richtext": "api-shapes:FormattedTextValue_ClassShape",
    }
    g = Graph()
    for object_type, shacl_shape in property_type_mapper.items():
        g += _construct_one_property_type_text_value(onto, object_type, shacl_shape)
    return g


def _construct_one_property_type_text_value(onto: Graph, gui_element: str, shacl_shape: str) -> Graph:
    query_s = """
    PREFIX owl: <http://www.w3.org/2002/07/owl#> 
    PREFIX sh: <http://www.w3.org/ns/shacl#>
    PREFIX api-shapes: <http://api.knora.org/ontology/knora-api/shapes/v2#>
    PREFIX knora-api:  <http://api.knora.org/ontology/knora-api/v2#>
    PREFIX salsah-gui: <http://api.knora.org/ontology/salsah-gui/v2#>

    CONSTRUCT {

        ?shapesIRI a sh:PropertyShape ;
                   sh:path ?prop ;
                   sh:node %(shacl_shape)s .

    } WHERE {

        ?prop a owl:ObjectProperty ;
                knora-api:objectType knora-api:TextValue ;
                salsah-gui:guiElement %(gui_element)s .

        BIND(IRI(CONCAT(str(?prop), "_PropShape")) AS ?shapesIRI)
    }
    """ % {"gui_element": gui_element, "shacl_shape": shacl_shape}  # noqa: UP031 (printf-string-formatting)
    if results_graph := onto.query(query_s).graph:
        return results_graph
    return Graph()



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib import Namespace

from dsp_tools.commands.validate_data.models.input_problems import ContentRegexViolation
from dsp_tools.commands.validate_data.models.input_problems import MaxCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import MinCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import NonExistentCardinalityViolation
from dsp_tools.commands.validate_data.models.input_problems import ValueTypeViolation
from dsp_tools.commands.validate_data.models.validation import CardinalityValidationResult
from dsp_tools.commands.validate_data.models.validation import ContentValidationResult
from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.commands.validate_data.reformat_validaton_result import _query_for_one_cardinality_validation_result
from dsp_tools.commands.validate_data.reformat_validaton_result import _query_for_one_content_validation_result
from dsp_tools.commands.validate_data.reformat_validaton_result import _reformat_one_cardinality_validation_result
from dsp_tools.commands.validate_data.reformat_validaton_result import _reformat_one_content_validation_result

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
DATA = Namespace("http://data/")
KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")


class TestQueryCardinality:
    def test_query_for_one_cardinality_validation_result(
        self, min_count_violation: Graph, data_min_count_violation: Graph
    ) -> None:
        bn = next(min_count_violation.subjects(RDF.type, SH.ValidationResult))
        result = _query_for_one_cardinality_validation_result(bn, min_count_violation, data_min_count_violation)
        assert result.source_constraint_component == SH.MinCountConstraintComponent
        assert result.res_iri == DATA.id_min_card
        assert result.res_class == ONTO.ClassMixedCard
        assert result.property == ONTO.testGeoname
        assert result.results_message == "1-n"


class TestReformatCardinalityViolation:
    def test_min(self, violation_min_card: CardinalityValidationResult) -> None:
        result = _reformat_one_cardinality_validation_result(violation_min_card)
        assert isinstance(result, MinCardinalityViolation)
        assert result.res_id == "id_min_card"
        assert result.res_type == "onto:ClassMixedCard"
        assert result.prop_name == "onto:testGeoname"
        assert result.expected_cardinality == "1-n"

    def test_max(self, violation_max_card: CardinalityValidationResult) -> None:
        result = _reformat_one_cardinality_validation_result(violation_max_card)
        assert isinstance(result, MaxCardinalityViolation)
        assert result.res_id == "id_max_card"
        assert result.res_type == "onto:ClassMixedCard"
        assert result.prop_name == "onto:testDecimalSimpleText"
        assert result.expected_cardinality == "0-1"

    def test_closed(self, violation_closed: CardinalityValidationResult) -> None:
        result = _reformat_one_cardinality_validation_result(violation_closed)
        assert isinstance(result, NonExistentCardinalityViolation)
        assert result.res_id == "id_closed_constraint"
        assert result.res_type == "onto:CardOneResource"
        assert result.prop_name == "onto:testIntegerSimpleText"


class TestQueryGraphContent:
    def test_value_type_mismatch(self, graph_value_type_mismatch: Graph, data_value_type_mismatch: Graph) -> None:
        bn = next(graph_value_type_mismatch.subjects(SH.sourceConstraintComponent, SH.NodeConstraintComponent))
        result = _query_for_one_content_validation_result(bn, graph_value_type_mismatch, data_value_type_mismatch)
        assert result.source_constraint_component == SH.NodeConstraintComponent
        assert result.res_iri == DATA.value_type_mismatch
        assert result.res_class == ONTO.ClassWithEverything
        assert result.property == ONTO.testColor
        assert result.results_message == "ColorValue"
        assert result.value_type == KNORA_API.TextValue

    def test_wrong_regex_content(self, graph_wrong_regex_content: Graph, data_wrong_regex_content: Graph) -> None:
        bn = next(graph_wrong_regex_content.subjects(SH.sourceConstraintComponent, SH.NodeConstraintComponent))
        result = _query_for_one_content_validation_result(bn, graph_wrong_regex_content, data_wrong_regex_content)
        assert result.source_constraint_component == SH.NodeConstraintComponent
        assert result.res_iri == DATA.geoname_not_number
        assert result.res_class == ONTO.ClassWithEverything
        assert result.property == ONTO.testGeoname
        assert result.results_message == "The value must be a valid geoname code"
        assert result.detail_bn_component == SH.PatternConstraintComponent
        assert result.value_type == KNORA_API.GeonameValue
        assert result.value == "this-is-not-a-valid-code"


class TestReformatContentViolation:
    def test_value_type(self, violation_value_type: ContentValidationResult) -> None:
        result = _reformat_one_content_validation_result(violation_value_type)
        assert isinstance(result, ValueTypeViolation)
        assert result.res_id == "id_2"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testColor"
        assert result.actual_type == "TextValue"
        assert result.expected_type == "ColorValue"

    def test_violation_regex(self, violation_regex: ContentValidationResult) -> None:
        result = _reformat_one_content_validation_result(violation_regex)
        assert isinstance(result, ContentRegexViolation)
        assert result.res_id == "geoname_not_number"
        assert result.res_type == "onto:ClassWithEverything"
        assert result.prop_name == "onto:testGeoname"
        assert result.expected_format == "The value must be a valid geoname code"
        assert result.actual_content == "this-is-not-a-valid-code"

    def test_unknown(self, violation_unknown_content: ContentValidationResult) -> None:
        result = _reformat_one_content_validation_result(violation_unknown_content)
        assert isinstance(result, UnexpectedComponent)
        assert result.component_type == str(SH.UniqueLangConstraintComponent)


if __name__ == "__main__":
    pytest.main([__file__])



from unittest.mock import Mock
from unittest.mock import patch

import pytest

from dsp_tools.commands.validate_data.api_connection import OntologyConnection
from dsp_tools.models.exceptions import UserError


@pytest.fixture
def ontology_connection() -> OntologyConnection:
    return OntologyConnection("http://0.0.0.0:3333", "9999")


def test_get_ontology_iris_ok(ontology_connection: OntologyConnection) -> None:
    mock_response = Mock()
    mock_response.ok = True
    mock_response.json.return_value = {"project": {"ontologies": ["onto_iri"]}}
    with patch.object(ontology_connection, "_get", return_value=mock_response) as patched_get:
        result = ontology_connection._get_ontology_iris()
        assert result == ["onto_iri"]
        patched_get.assert_called_once_with("http://0.0.0.0:3333/admin/projects/shortcode/9999")


def test_get_ontology_iris_non_ok_code(ontology_connection: OntologyConnection) -> None:
    mock_response = Mock()
    mock_response.ok = False
    mock_response.json.return_value = {}
    with patch.object(ontology_connection, "_get", return_value=mock_response) as patched_get:
        with pytest.raises(UserError):
            ontology_connection._get_ontology_iris()
        patched_get.assert_called_once_with("http://0.0.0.0:3333/admin/projects/shortcode/9999")


def test_get_ontology_iris_no_ontology_key(ontology_connection: OntologyConnection) -> None:
    mock_response = Mock()
    mock_response.ok = True
    mock_response.json.return_value = {}
    with patch.object(ontology_connection, "_get", return_value=mock_response) as patched_get:
        with pytest.raises(UserError):
            ontology_connection._get_ontology_iris()
        patched_get.assert_called_once_with("http://0.0.0.0:3333/admin/projects/shortcode/9999")


def test_get_one_ontology(ontology_connection: OntologyConnection) -> None:
    mock_response = Mock()
    mock_response.ok = True
    mock_response.text = "Turtle Text"
    with patch.object(ontology_connection, "_get", return_value=mock_response) as patched_get:
        result = ontology_connection._get_one_ontology("iri")
        assert result == "Turtle Text"
        patched_get.assert_called_once_with("iri", headers={"Accept": "text/turtle"})


if __name__ == "__main__":
    pytest.main([__file__])



from typing import Any

import pandas as pd
import pytest

from dsp_tools.xmllib.models.user_enums import NewlineReplacement
from dsp_tools.xmllib.value_converters import convert_to_bool_string
from dsp_tools.xmllib.value_converters import replace_newlines_with_tags


@pytest.mark.parametrize("val", ["false", "0  ", "  0.0", "NO", False])
def test_convert_to_bool_false(val: Any) -> None:
    assert convert_to_bool_string(val) == "false"


@pytest.mark.parametrize("val", ["TRUE ", "  1", "1.0", "Yes", True])
def test_convert_to_bool_true(val: Any) -> None:
    assert convert_to_bool_string(val) == "true"


@pytest.mark.parametrize("val", [pd.NA, None, 2.421, 10, "other", "  "])
def test_convert_to_bool_failure(val: Any) -> None:
    assert convert_to_bool_string(val) == str(val)


def test_replace_newlines_with_tags_none() -> None:
    text = "Start\nMiddle\n\nFinal"
    result = replace_newlines_with_tags(text, NewlineReplacement.NONE)
    assert result == text


def test_replace_newlines_with_tags_newline() -> None:
    text = "Start\nMiddle\n\nFinal"
    result = replace_newlines_with_tags(text, NewlineReplacement.LINEBREAK)
    assert result == "Start<br/>Middle<br/><br/>Final"


def test_replace_newlines_with_tags_paragraph() -> None:
    text = "Start\nMiddle\n\nFinal"
    result = replace_newlines_with_tags(text, NewlineReplacement.PARAGRAPH)
    assert result == "<p>Start</p><p>Middle</p><p>Final</p>"


if __name__ == "__main__":
    pytest.main([__file__])



import warnings
from typing import TextIO

from dsp_tools.models.custom_warnings import DspToolsWarning


def initialize_warnings() -> None:
    """
    This function makes sure that DSP-TOOLS internal warnings are displayed in their custom way how they specify it.
    This is done by monkeypatching the behavior of the warnings module, as officially recommended by the Python docs:
    https://docs.python.org/3/library/warnings.html#warnings.showwarning
    """

    built_in_showwarning = warnings.showwarning

    def _custom_showwarning(
        message: Warning | str,
        category: type[Warning],
        filename: str,
        lineno: int,
        file: TextIO | None = None,
        line: str | None = None,
    ) -> None:
        if issubclass(category, DspToolsWarning):
            category.showwarning(str(message))
        else:
            built_in_showwarning(message, category, filename, lineno, file, line)

    warnings.showwarning = _custom_showwarning



from pathlib import Path

import pytest
import regex

from dsp_tools.commands.excel2xml import excel2xml_cli
from dsp_tools.models.exceptions import BaseError

INVALID_EXCEL_DIRECTORY = "testdata/invalid-testdata/excel2xml"


@pytest.fixture
def expected_output() -> str:
    with open("testdata/excel2xml/excel2xml-expected-output.xml", encoding="utf-8") as f:
        return f.read()


class TestDifferentExtensions:
    def test_xlsx(self, expected_output: str) -> None:
        success, problems = excel2xml_cli.excel2xml(
            "testdata/excel2xml/excel2xml-testdata.xlsx", "1234", "excel2xml-output"
        )
        assert success
        assert not problems
        returned = Path("excel2xml-output-data.xml")
        assert returned.read_text() == expected_output
        returned.unlink(missing_ok=True)

    def test_xls(self, expected_output: str) -> None:
        success, problems = excel2xml_cli.excel2xml(
            "testdata/excel2xml/excel2xml-testdata.xls", "1234", "excel2xml-output"
        )
        assert success
        assert not problems
        returned = Path("excel2xml-output-data.xml")
        assert returned.read_text() == expected_output
        returned.unlink(missing_ok=True)

    def test_csv(self, expected_output: str) -> None:
        success, problems = excel2xml_cli.excel2xml(
            "testdata/excel2xml/excel2xml-testdata.csv", "1234", "excel2xml-output"
        )
        assert success
        assert not problems
        returned = Path("excel2xml-output-data.xml")
        assert returned.read_text() == expected_output
        returned.unlink(missing_ok=True)


class TestWarnings:
    def test_double_bool(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/boolean-prop-two-values.xlsx"
        expected_msg = (
            "A <boolean-prop> can only have a single value, but resource "
            "'test_thing_1', property ':hasBoolean' (Excel row 6) contains more than one value."
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 1
        catched = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert catched == expected_msg

    def test_missing_child(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/empty-property.xlsx"
        expected_missing_prop = regex.escape(
            "At least one value per property is required, but resource 'person_0', property ':hasName' "
            "(Excel row 3) doesn't contain any values."
        )
        expected_validation = regex.escape(
            "Line 28: Element 'text-prop': Missing child element(s). Expected is ( text )."
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 2
        message_missing_prop = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_missing_prop, message_missing_prop)
        message_xml_validation = catched_warnings[1].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_validation, message_xml_validation)

    def test_missing_prop_permission(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/missing-prop-permissions.xlsx"
        expected_msg = "Resource 'person_0': Missing permissions in column '2_permissions' of property ':hasName'"
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 1
        message = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert message == expected_msg

    def test_missing_label(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/missing-resource-label.xlsx"
        expected_msg_missing = "Missing label for resource 'person_0' (Excel row 2)"
        expected_xml_validation = regex.escape(
            "Line 27: Element 'resource', attribute 'label': [facet 'minLength'] The value '' has a length of '0'"
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 2
        msg_missing_label = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert msg_missing_label == expected_msg_missing
        msg_validation = catched_warnings[1].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_xml_validation, msg_validation)

    def test_missing_resource_permission(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/missing-resource-permissions.xlsx"
        expected_msg = "Missing permissions for resource 'person_0' (Excel row 2)"
        expected_xml_validation = regex.escape(
            "Line 27: Element 'resource', attribute 'permissions': "
            "'' is not a valid value of the atomic type 'xs:NCName'."
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 2
        message = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert message == expected_msg
        msg_validation = catched_warnings[1].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_xml_validation, msg_validation)

    def test_missing_restype(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/missing-restype.xlsx"
        expected_msg = "Missing restype for resource 'person_0' (Excel row 2)"
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 1
        message = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert message == expected_msg

    def test_missing_bitstream_permission(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/no-bitstream-permissions.xlsx"
        expected_msg_missing = (
            "Missing file permissions for file 'testdata/bitstreams/test.jpg' "
            "(Resource ID 'test_thing_1', Excel row 2)."
            " An attempt to deduce them from the resource permissions failed."
        )
        expected_xml_validation = regex.escape(
            "Line 28: Element 'bitstream', attribute 'permissions': "
            "'' is not a valid value of the atomic type 'xs:NCName'."
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 2
        msg_missing_label = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert msg_missing_label == expected_msg_missing
        msg_validation = catched_warnings[1].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_xml_validation, msg_validation)

    def test_invalid_prop_val(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/single-invalid-value-for-property.xlsx"
        expected_msg_missing = regex.escape(
            "Error in resource 'person_0': Excel row 3 has an entry in column(s) '1_encoding', '1_permissions', "
            "but not in '1_value'. Please note that cell contents that don't meet the requirements"
        )
        expected_msg_excel = (
            "At least one value per property is required, "
            "but resource 'person_0', property ':hasName' (Excel row 3) doesn't contain any values."
        )
        expected_xml_validation = regex.escape(
            "Line 28: Element 'text-prop': Missing child element(s). Expected is ( text )."
        )
        _, catched_warnings = excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")
        assert len(catched_warnings) == 3
        msg_missing_label = catched_warnings[0].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_msg_missing, msg_missing_label)
        msg_excel = catched_warnings[1].message.args[0]  # type:ignore[union-attr]
        assert msg_excel == expected_msg_excel
        msg_validation = catched_warnings[2].message.args[0]  # type:ignore[union-attr]
        assert regex.search(expected_xml_validation, msg_validation)


class TestRaisesException:
    def test_id_prop_name(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/id-propname-both.xlsx"
        expected_msg = regex.escape("Exactly 1 of the 2 columns 'id' and 'prop name' must be filled.")
        with pytest.raises(BaseError, match=expected_msg):
            excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")

    def test_id_prop_none(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/id-propname-none.xlsx"
        expected_msg = regex.escape("Exactly 1 of the 2 columns 'id' and 'prop name' must be filled.")
        with pytest.raises(BaseError, match=expected_msg):
            excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")

    def test_nonexisting_proptype(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/nonexisting-proptype.xlsx"
        expected_msg = regex.escape("Invalid prop type for property :hasName in resource person_0")
        with pytest.raises(BaseError, match=expected_msg):
            excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")

    def test_start_with_prop_row(self) -> None:
        file = f"{INVALID_EXCEL_DIRECTORY}/start-with-property-row.xlsx"
        expected_msg = regex.escape(
            "The first row of your Excel/CSV is invalid. The first row must define a resource, not a property."
        )
        with pytest.raises(BaseError, match=expected_msg):
            excel2xml_cli.excel2xml(file, "1234", "excel2xml-invalid")


if __name__ == "__main__":
    pytest.main([__file__])



import json
from typing import Any
from typing import Union

from dsp_tools.commands.project.models.context import Context
from dsp_tools.commands.project.models.helpers import OntoIri


class SetEncoder(json.JSONEncoder):
    """Encoder used to serialize objects to JSON that would by default not be serializable"""

    def default(self, o: Union[set[Any], Context, OntoIri]) -> Any:
        """Return a serializable object for o"""
        if isinstance(o, set):
            return list(o)
        elif isinstance(o, Context):
            return o.toJsonObj()
        elif isinstance(o, OntoIri):
            return {"iri": o.iri, "hashtag": o.hashtag}
        return json.JSONEncoder.default(self, o)



from typing import Callable

from rdflib import XSD
from rdflib import Literal
from rdflib import Namespace
from rdflib import URIRef

from dsp_tools.commands.validate_data.models.data_deserialised import AbstractResource
from dsp_tools.commands.validate_data.models.data_deserialised import AnnotationDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import AudioSegmentDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import BooleanValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ColorValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DataDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DateValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import DecimalValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import GeonameValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import IntValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import LinkObjDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import LinkValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ListValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import RegionDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ResourceDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import RichtextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import SimpleTextDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import TimeValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import UriValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import ValueDeserialised
from dsp_tools.commands.validate_data.models.data_deserialised import VideoSegmentDeserialised
from dsp_tools.commands.validate_data.models.data_rdf import BooleanValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ColorValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import DataRDF
from dsp_tools.commands.validate_data.models.data_rdf import DateValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import DecimalValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import GeonameValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import IntValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import LinkValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ListValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import RDFTriples
from dsp_tools.commands.validate_data.models.data_rdf import ResourceRDF
from dsp_tools.commands.validate_data.models.data_rdf import RichtextRDF
from dsp_tools.commands.validate_data.models.data_rdf import SimpleTextRDF
from dsp_tools.commands.validate_data.models.data_rdf import TimeValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import UriValueRDF
from dsp_tools.commands.validate_data.models.data_rdf import ValueRDF
from dsp_tools.models.exceptions import InternalError

KNORA_API = Namespace("http://api.knora.org/ontology/knora-api/v2#")
DATA = Namespace("http://data/")


def make_data_rdf(data_deserialised: DataDeserialised) -> DataRDF:
    """
    Transforms the deserialised data into instances that can produce a RDF graph.

    Args:
        data_deserialised: Deserialised Data

    Returns:
        Instance with the data
    """
    all_triples: list[RDFTriples] = []
    for r in data_deserialised.resources:
        all_triples.extend(_transform_one_resource(r))
    return DataRDF(all_triples)


def _transform_one_resource(res: AbstractResource) -> list[RDFTriples]:
    if isinstance(res, ResourceDeserialised):
        return _transform_one_project_resource(res)
    return [_transform_one_dsp_resource(res)]


def _transform_one_dsp_resource(res: AbstractResource) -> RDFTriples:
    res_type_mapper = {
        AnnotationDeserialised: KNORA_API.Annotation,
        RegionDeserialised: KNORA_API.Region,
        LinkObjDeserialised: KNORA_API.LinkObj,
        VideoSegmentDeserialised: KNORA_API.VideoSegment,
        AudioSegmentDeserialised: KNORA_API.AudioSegment,
    }
    return ResourceRDF(
        res_iri=DATA[res.res_id], res_class=res_type_mapper[type(res)], label=Literal(res.label, datatype=XSD.string)
    )


def _transform_one_project_resource(res: ResourceDeserialised) -> list[RDFTriples]:
    res_iri = DATA[res.res_id]
    all_triples: list[RDFTriples] = [
        ResourceRDF(res_iri=res_iri, res_class=URIRef(res.res_class), label=Literal(res.label, datatype=XSD.string))
    ]
    all_triples.extend([_transform_one_value(v, res_iri) for v in res.values])
    return all_triples


def _transform_one_value(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:  # noqa: PLR0911 (too many return statements)
    func_mapper = {
        ColorValueDeserialised: ColorValueRDF,
        DateValueDeserialised: DateValueRDF,
        GeonameValueDeserialised: GeonameValueRDF,
        IntValueDeserialised: IntValueRDF,
        SimpleTextDeserialised: SimpleTextRDF,
        RichtextDeserialised: RichtextRDF,
    }
    match val:
        case (
            ColorValueDeserialised()
            | DateValueDeserialised()
            | GeonameValueDeserialised()
            | SimpleTextDeserialised()
            | RichtextDeserialised()
        ):
            return _transform_into_xsd_string(val, res_iri, func_mapper[type(val)])
        case IntValueDeserialised():
            return _transform_into_xsd_integer(val, res_iri, func_mapper[type(val)])
        case BooleanValueDeserialised():
            return _transform_into_bool(val, res_iri)
        case DecimalValueDeserialised():
            return _transform_decimal_value(val, res_iri)
        case LinkValueDeserialised():
            return _transform_link_value(val, res_iri)
        case ListValueDeserialised():
            return _transform_list_value(val, res_iri)
        case TimeValueDeserialised():
            return _transform_time_value(val, res_iri)
        case UriValueDeserialised():
            return _transform_uri_value(val, res_iri)
        case _:
            raise InternalError(f"Unknown Value Type: {type(val)}")


def _transform_into_xsd_string(
    val: ValueDeserialised, res_iri: URIRef, func: Callable[[URIRef, Literal, URIRef], ValueRDF]
) -> ValueRDF:
    new_str = val.object_value if val.object_value is not None else ""
    return func(URIRef(val.prop_name), Literal(new_str, datatype=XSD.string), res_iri)


def _transform_into_xsd_integer(
    val: ValueDeserialised, res_iri: URIRef, func: Callable[[URIRef, Literal, URIRef], ValueRDF]
) -> ValueRDF:
    content = (
        Literal(val.object_value, datatype=XSD.integer)
        if val.object_value is not None
        else Literal("", datatype=XSD.string)
    )
    return func(URIRef(val.prop_name), content, res_iri)


def _transform_into_bool(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:
    match val.object_value:
        case "1" | "true":
            content = Literal(True, datatype=XSD.boolean)
        case "0" | "false":
            content = Literal(False, datatype=XSD.boolean)
        case _:
            content = Literal("", datatype=XSD.string)
    return BooleanValueRDF(URIRef(val.prop_name), content, res_iri)


def _transform_decimal_value(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:
    content = (
        Literal(val.object_value, datatype=XSD.decimal)
        if val.object_value is not None
        else Literal("", datatype=XSD.string)
    )
    return DecimalValueRDF(URIRef(val.prop_name), content, res_iri)


def _transform_link_value(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:
    content = val.object_value if val.object_value is not None else ""
    return LinkValueRDF(URIRef(val.prop_name), DATA[content], res_iri)


def _transform_list_value(val: ListValueDeserialised, res_iri: URIRef) -> ValueRDF:
    node_name = val.object_value if val.object_value is not None else ""
    return ListValueRDF(
        prop_name=URIRef(val.prop_name),
        object_value=Literal(node_name, datatype=XSD.string),
        list_name=Literal(val.list_name, datatype=XSD.string),
        res_iri=res_iri,
    )


def _transform_time_value(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:
    content = (
        Literal(val.object_value, datatype=XSD.dateTimeStamp)
        if val.object_value is not None
        else Literal("", datatype=XSD.string)
    )
    return TimeValueRDF(URIRef(val.prop_name), content, res_iri)


def _transform_uri_value(val: ValueDeserialised, res_iri: URIRef) -> ValueRDF:
    content = (
        Literal(val.object_value, datatype=XSD.anyURI)
        if val.object_value is not None
        else Literal("", datatype=XSD.string)
    )
    return UriValueRDF(URIRef(val.prop_name), content, res_iri)



import shutil
from typing import Iterator

import pytest

from test.e2e.setup_testcontainers import INGEST_DB
from test.e2e.setup_testcontainers import SIPI_IMAGES
from test.e2e.setup_testcontainers import TMP_INGEST
from test.e2e.setup_testcontainers import TMP_SIPI


@pytest.fixture(scope="package", autouse=True)
def _tidy_up_artifacts() -> Iterator[None]:
    """Tidy up artifacts after each test, to prevent cluttering the local clone of the repo."""
    yield
    shutil.rmtree(SIPI_IMAGES, ignore_errors=True)
    shutil.rmtree(TMP_INGEST, ignore_errors=True)
    shutil.rmtree(TMP_SIPI, ignore_errors=True)
    shutil.rmtree(INGEST_DB, ignore_errors=True)



import pandas as pd
import pytest
import regex

from dsp_tools.models.exceptions import InputError
from dsp_tools.xmllib.models.file_values import FileValue
from dsp_tools.xmllib.models.file_values import IIIFUri
from dsp_tools.xmllib.models.resource import Resource
from dsp_tools.xmllib.models.user_enums import NewlineReplacement
from dsp_tools.xmllib.models.values import BooleanValue
from dsp_tools.xmllib.models.values import ColorValue
from dsp_tools.xmllib.models.values import DateValue
from dsp_tools.xmllib.models.values import DecimalValue
from dsp_tools.xmllib.models.values import GeonameValue
from dsp_tools.xmllib.models.values import IntValue
from dsp_tools.xmllib.models.values import LinkValue
from dsp_tools.xmllib.models.values import ListValue
from dsp_tools.xmllib.models.values import Richtext
from dsp_tools.xmllib.models.values import SimpleText
from dsp_tools.xmllib.models.values import TimeValue
from dsp_tools.xmllib.models.values import UriValue


class TestAddValues:
    def test_add_bool(self) -> None:
        res = Resource("", "", "").add_bool("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], BooleanValue)

    def test_add_bools(self) -> None:
        res = Resource("", "", "").add_bools(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, BooleanValue) for x in res.values])

    def test_add_bool_optional(self) -> None:
        res = Resource("", "", "").add_bool_optional(None, "")
        assert not res.values
        res = res.add_bool_optional("val", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], BooleanValue)

    def test_add_color(self) -> None:
        res = Resource("", "", "").add_color("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ColorValue)

    def test_add_colors(self) -> None:
        res = Resource("", "", "").add_colors(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, ColorValue) for x in res.values])

    def test_add_color_optional(self) -> None:
        res = Resource("", "", "").add_color_optional(pd.NA, "")
        assert not res.values
        res = res.add_color_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ColorValue)

    def test_add_date(self) -> None:
        res = Resource("", "", "").add_date("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DateValue)

    def test_add_dates(self) -> None:
        res = Resource("", "", "").add_dates(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, DateValue) for x in res.values])

    def test_add_date_optional(self) -> None:
        res = Resource("", "", "").add_date_optional(None, "")
        assert not res.values
        res = res.add_date_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DateValue)

    def test_add_decimal(self) -> None:
        res = Resource("", "", "").add_decimal("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DecimalValue)

    def test_add_decimals(self) -> None:
        res = Resource("", "", "").add_decimals(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, DecimalValue) for x in res.values])

    def test_add_decimal_optional(self) -> None:
        res = Resource("", "", "").add_decimal_optional(None, "")
        assert not res.values
        res = res.add_decimal_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], DecimalValue)

    def test_add_geoname(self) -> None:
        res = Resource("", "", "").add_geoname("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], GeonameValue)

    def test_add_geonames(self) -> None:
        res = Resource("", "", "").add_geonames(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, GeonameValue) for x in res.values])

    def test_add_geoname_optional(self) -> None:
        res = Resource("", "", "").add_geoname_optional(pd.NA, "")
        assert not res.values
        res = res.add_geoname_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], GeonameValue)

    def test_add_integer(self) -> None:
        res = Resource("", "", "").add_integer("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], IntValue)

    def test_add_integers(self) -> None:
        res = Resource("", "", "").add_integers(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, IntValue) for x in res.values])

    def test_add_integer_optional(self) -> None:
        res = Resource("", "", "").add_integer_optional(None, "")
        assert not res.values
        res = res.add_integer_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], IntValue)

    def test_add_link(self) -> None:
        res = Resource("", "", "").add_link("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], LinkValue)

    def test_add_links(self) -> None:
        res = Resource("", "", "").add_links(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, LinkValue) for x in res.values])

    def test_add_link_optional(self) -> None:
        res = Resource("", "", "").add_link_optional(pd.NA, "")
        assert not res.values
        res = res.add_link_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], LinkValue)

    def test_add_list(self) -> None:
        res = Resource("", "", "").add_list("", "", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ListValue)

    def test_add_lists(self) -> None:
        res = Resource("", "", "").add_lists(["", ""], "", "")
        assert len(res.values) == 2
        assert all([isinstance(x, ListValue) for x in res.values])

    def test_add_list_optional(self) -> None:
        res = Resource("", "", "").add_list_optional(None, None, "")
        assert not res.values
        res = res.add_list_optional("", "", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], ListValue)

    def test_add_simple_text(self) -> None:
        res = Resource("", "", "").add_simpletext("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], SimpleText)

    def test_add_simple_texts(self) -> None:
        res = Resource("", "", "").add_simpletexts(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, SimpleText) for x in res.values])

    def test_add_simple_text_optional(self) -> None:
        res = Resource("", "", "").add_simpletext_optional(None, "")
        assert not res.values
        res = res.add_simpletext_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], SimpleText)

    def test_add_richtext(self) -> None:
        res = Resource("", "", "").add_richtext("A\nB", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)
        assert res.values[0].value == "A<br/>B"

    def test_add_richtext_no_replace(self) -> None:
        res = Resource("", "", "").add_richtext("A\nB", "", newline_replacement=NewlineReplacement.NONE)
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)
        assert res.values[0].value == "A\nB"

    def test_add_richtexts(self) -> None:
        res = Resource("", "", "").add_richtexts(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, Richtext) for x in res.values])

    def test_add_richtext_optional(self) -> None:
        res = Resource("", "", "").add_richtext_optional(None, "")
        assert not res.values
        res = res.add_richtext_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], Richtext)

    def test_add_time(self) -> None:
        res = Resource("", "", "").add_time("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], TimeValue)

    def test_add_times(self) -> None:
        res = Resource("", "", "").add_times(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, TimeValue) for x in res.values])

    def test_add_time_optional(self) -> None:
        res = Resource("", "", "").add_time_optional(None, "")
        assert not res.values
        res = res.add_time_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], TimeValue)

    def test_add_uri(self) -> None:
        res = Resource("", "", "").add_uri("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], UriValue)

    def test_add_uris(self) -> None:
        res = Resource("", "", "").add_uris(["", ""], "")
        assert len(res.values) == 2
        assert all([isinstance(x, UriValue) for x in res.values])

    def test_add_uri_optional(self) -> None:
        res = Resource("", "", "").add_uri_optional(pd.NA, "")
        assert not res.values
        res = res.add_uri_optional("", "")
        assert len(res.values) == 1
        assert isinstance(res.values[0], UriValue)


class TestAddFiles:
    def test_add_file(self) -> None:
        res = Resource("", "", "").add_file("")
        assert isinstance(res.file_value, FileValue)

    def test_add_file_raising(self) -> None:
        res = Resource("id", "", "").add_file("existing filename")
        msg = regex.escape(
            "The resource with the ID 'id' already contains a file with the name: 'existing filename'.\n"
            "The new file with the name 'new filename' cannot be added."
        )
        with pytest.raises(InputError, match=msg):
            res.add_file("new filename")

    def test_add_iiif_uri(self) -> None:
        res = Resource("", "", "").add_iiif_uri("")
        assert isinstance(res.file_value, IIIFUri)

    def test_add_iiif_uri_raising(self) -> None:
        res = Resource("id", "", "").add_file("existing IIIF")
        msg = regex.escape(
            "The resource with the ID 'id' already contains a file with the name: 'existing IIIF'.\n"
            "The new file with the name 'new IIIF' cannot be added."
        )
        with pytest.raises(InputError, match=msg):
            res.add_iiif_uri("new IIIF")


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

from typing import Any
from typing import cast

import regex
import rustworkx as rx
from lxml import etree

from dsp_tools.commands.xmlupload.stash.graph_models import Cost
from dsp_tools.commands.xmlupload.stash.graph_models import Edge
from dsp_tools.commands.xmlupload.stash.graph_models import ResptrLink
from dsp_tools.commands.xmlupload.stash.graph_models import XMLLink
from dsp_tools.utils.iri_util import is_resource_iri


def create_info_from_xml_for_graph(
    root: etree._Element,
) -> tuple[list[ResptrLink], list[XMLLink], list[str]]:
    """
    Create link objects (ResptrLink/XMLLink) from the XML file,
    and add a reference UUID to each XML element that contains a link (<resptr> or <text>).
    With this UUID, the link objects can be identified in the XML data file.

    Args:
        root: root of the parsed XML file

    Returns:
        - All resptr links contained in the XML file, represented as ResptrLink objects.
        - All XML links contained in the XML file, represented as XMLLink objects.
        - A list with all resource IDs used in the XML file.
    """
    resptr_links = []
    xml_links = []
    all_resource_ids = []
    for resource in root.iter(tag="resource"):
        resptr, xml = _create_info_from_xml_for_graph_from_one_resource(resource)
        all_resource_ids.append(resource.attrib["id"])
        resptr_links.extend(resptr)
        xml_links.extend(xml)
    return resptr_links, xml_links, all_resource_ids


def _create_info_from_xml_for_graph_from_one_resource(
    resource: etree._Element,
) -> tuple[list[ResptrLink], list[XMLLink]]:
    resptr_links: list[ResptrLink] = []
    xml_links: list[XMLLink] = []
    for prop in resource.getchildren():
        match prop.tag:
            case "resptr-prop":
                resptr_links.extend(_create_resptr_link_objects(resource.attrib["id"], prop))
            case "text-prop":
                xml_links.extend(_create_text_link_objects(resource.attrib["id"], prop))
            case "hasComment" | "hasDescription":
                if xml_link := _create_text_link_object_from_special_tags(resource.attrib["id"], prop):
                    xml_links.append(xml_link)
            case "isSegmentOf" | "relatesTo":
                if segment_link := _create_segmentOf_link_objects(resource.attrib["id"], prop):
                    resptr_links.append(segment_link)
    return resptr_links, xml_links


def _create_segmentOf_link_objects(subject_id: str, resptr: etree._Element) -> ResptrLink | None:
    resptr.text = cast(str, resptr.text)
    if is_resource_iri(resptr.text):
        return None
    link_object = ResptrLink(subject_id, resptr.text)
    # this UUID is so that the links that were stashed can be identified in the XML data file
    resptr.attrib["linkUUID"] = link_object.link_uuid
    return link_object


def _create_resptr_link_objects(subject_id: str, resptr_prop: etree._Element) -> list[ResptrLink]:
    resptr_links = []
    for resptr in resptr_prop.getchildren():
        resptr.text = cast(str, resptr.text)
        if not is_resource_iri(resptr.text):
            link_object = ResptrLink(subject_id, resptr.text)
            # this UUID is so that the links that were stashed can be identified in the XML data file
            resptr.attrib["linkUUID"] = link_object.link_uuid
            resptr_links.append(link_object)
    return resptr_links


def _create_text_link_objects(subject_id: str, text_prop: etree._Element) -> list[XMLLink]:
    # if the same ID is in several separate <text> values of one <text-prop>, they are considered separate links
    xml_props = []
    for text in text_prop.getchildren():
        if links := _extract_ids_from_one_text_value(text):
            xml_link = XMLLink(subject_id, links)
            xml_props.append(xml_link)
            # this UUID is so that the links that were stashed can be identified in the XML data file
            text.attrib["linkUUID"] = xml_link.link_uuid
    return xml_props


def _create_text_link_object_from_special_tags(subject_id: str, special_tag: etree._Element) -> XMLLink | None:
    # This is for <hasDescription> and <hasComment> properties of <video-segment>s or <audio-segment>s
    if not (links := _extract_ids_from_one_text_value(special_tag)):
        return None
    xml_link = XMLLink(subject_id, links)
    # this UUID is so that the links that were stashed can be identified in the XML data file
    special_tag.attrib["linkUUID"] = xml_link.link_uuid
    return xml_link


def _extract_ids_from_one_text_value(text: etree._Element) -> set[str]:
    # the same id in one <text> only means one link to the resource
    all_links = set()
    for ele in text.iterdescendants():
        if href := ele.attrib.get("href"):
            if internal_id := regex.search(r"IRI:(.*):IRI", href):
                all_links.add(internal_id.group(1))
    return all_links


def make_graph(
    resptr_links: list[ResptrLink],
    xml_links: list[XMLLink],
    all_resource_ids: list[str],
) -> tuple[rx.PyDiGraph[Any, Any], dict[int, str], list[Edge]]:
    """
    This function takes information about the resources of an XML file and links between them.
    From that it constructs a rustworkx directed graph.
    Resources are represented as nodes and links as edges.

    Args:
        resptr_links: objects representing a direct link between a starting resource and a target resource
        xml_links: objects representing one or more links from a single text value of a single starting resource
                   to a set of target resources
        all_resource_ids: IDs of all resources in the graph

    Returns:
        - The rustworkx graph.
        - A dictionary that maps the rustworkx index number of the nodes to the original resource ID from the XML file.
        - A list with all the edges in the graph.
    """
    graph: rx.PyDiGraph[Any, Any] = rx.PyDiGraph()
    nodes = [(id_, None, None) for id_ in all_resource_ids]
    node_indices = list(graph.add_nodes_from(nodes))
    id_to_node = dict(zip(all_resource_ids, node_indices))
    node_to_id = dict(zip(node_indices, all_resource_ids))
    edges = [Edge(id_to_node[x.source_id], id_to_node[x.target_id], x) for x in resptr_links]
    for xml in xml_links:
        edges.extend([Edge(id_to_node[xml.source_id], id_to_node[x], xml) for x in xml.target_ids])
    graph.add_edges_from([e.as_tuple() for e in edges])
    return graph, node_to_id, edges


def _remove_leaf_nodes(
    graph: rx.PyDiGraph[Any, Any],
    node_to_id: dict[int, str],
    node_indices: set[int],
) -> tuple[list[str], set[int]]:
    """
    Leaf nodes are nodes that do not have any outgoing links.
    This means that they have no dependencies and are ok to upload.
    This function removes them from the graph.

    Args:
        graph: graph
        node_to_id: mapping of the rustworkx index number of the nodes to the original resource ID from the XML file
        node_indices: node indices that are in the graph

    Returns:
        - A list with the IDs of the removed leaf nodes.
        - A set with the indices of the nodes that remain in the graph.
    """
    removed_leaf_nodes: list[str] = []
    remaining_node_indices = set(node_indices)
    while leaf_nodes := [x for x in remaining_node_indices if graph.out_degree(x) == 0]:
        removed_leaf_nodes.extend(node_to_id[n] for n in leaf_nodes)
        graph.remove_nodes_from(leaf_nodes)
        remaining_node_indices = remaining_node_indices - set(leaf_nodes)
    return removed_leaf_nodes, remaining_node_indices


def _find_cheapest_outgoing_links(
    graph: rx.PyDiGraph[Any, Any],
    cycle: list[tuple[int, int]],
    edges: list[Edge],
) -> list[Edge]:
    """
    This function searches for the nodes whose outgoing links should be removed in order to break the cycle.
    It calculates which links between the resources create the smallest stash.

    Args:
        graph: graph
        cycle: the list with (source, target) for each edge in the cycle
        edges: edges in the graph (contains info about source node, target node, and link info)

    Returns:
        The edges (i.e. links) that should be stashed (containing all the edges connecting the two nodes)
    """
    costs: list[Cost] = []
    for source, target in cycle:
        edges_in = graph.in_edges(source)
        node_gain = len(edges_in)
        edges_out = graph.out_edges(source)
        node_cost = sum(x[2].cost_links for x in edges_out)
        node_value = node_cost / node_gain
        costs.append(Cost(source, target, node_value))
    cheapest_cost = sorted(costs, key=lambda x: x.node_value)[0]
    return [x for x in edges if x.source == cheapest_cost.source and x.target == cheapest_cost.target]


def _remove_edges_to_stash(
    graph: rx.PyDiGraph[Any, Any],
    edges_to_remove: list[Edge],
    all_edges: list[Edge],
    remaining_nodes: set[int],
) -> None:
    """
    This function removes the edges from the graph in order to break a cycle.

    Args:
        graph: graph
        edges_to_remove: edges that should be removed
        all_edges: all edges in the original graph
        remaining_nodes: indices of the nodes in the graph
    """
    normal_edges_to_remove = [(x.source, x.target) for x in edges_to_remove]
    # if only one (source, target) is removed, it removes only one edge, not all
    # therefore we need as many entries in the list as there are edges between the source and the target

    phantom_edges_to_remove = []
    source, target = edges_to_remove[0].source, edges_to_remove[0].target
    for link_to_stash in [x.link_object for x in edges_to_remove]:
        if isinstance(link_to_stash, XMLLink):
            phantom_edges_to_remove.extend(
                _find_phantom_xml_edges(source, target, all_edges, link_to_stash, remaining_nodes)
            )

    all_edges_to_remove = normal_edges_to_remove + phantom_edges_to_remove
    graph.remove_edges_from(all_edges_to_remove)


def _find_phantom_xml_edges(
    source_node_index: int,
    target_node_index: int,
    all_edges: list[Edge],
    xml_link_to_stash: XMLLink,
    remaining_nodes: set[int],
) -> list[tuple[int, int]]:
    """
    If an edge that will be removed represents an XML link,
    the text value may contain further links to other resources.
    If we stash the XMLLink, then in the real data all links of that text value are stashed.
    So, these "phantom" links must be removed from the graph.
    This function identifies the edges that must be removed from the rx graph.

    Args:
        source_node_index: rustworkx index of source node
        target_node_index: rustworkx index of target node
        all_edges: all edges in the original graph
        xml_link_to_stash: XML link that will be stashed
        remaining_nodes: indices of all nodes in the graph

    Returns:
        edges (rustworkx indices of nodes) that represent the links in the original XML text
    """

    def check(x: Edge) -> bool:
        return all(
            (
                x.source == source_node_index,
                x.target != target_node_index,
                x.link_object == xml_link_to_stash,
                x.target in remaining_nodes,
                # the target could have been removed because it was a leaf node, so we must check if it is still there
            )
        )

    return [(x.source, x.target) for x in all_edges if check(x)]


def _add_stash_to_lookup_dict(
    stash_dict: dict[str, list[str]],
    links_to_stash: list[XMLLink | ResptrLink],
) -> dict[str, list[str]]:
    stash_list = [stash_link.link_uuid for stash_link in links_to_stash]
    # all stashed links have the same subject id, so we can just take the first one
    subj_id = links_to_stash[0].source_id
    if subj_id in stash_dict:
        stash_dict[subj_id].extend(stash_list)
    else:
        stash_dict[subj_id] = stash_list
    return stash_dict


def generate_upload_order(
    graph: rx.PyDiGraph[Any, Any],
    node_to_id: dict[int, str],
    edges: list[Edge],
) -> tuple[dict[str, list[str]], list[str], int]:
    """
    Generate the order in which the resources should be uploaded to the DSP-API based on the dependencies.

    Args:
        graph: graph
        node_to_id: mapping between indices of the graph nodes and original resource IDs from the XML file
        edges: edges in the graph (contains info about source node, target node, and link info)

    Returns:
        - A dictionary which maps the resources that have stashes to the UUIDs of the stashed links.
        - A list of resource IDs which gives the order in which the resources should be uploaded to DSP-API.
        - The number of links in the stash.
    """
    upload_order: list[str] = []
    stash_lookup: dict[str, list[str]] = {}
    node_indices = set(node_to_id.keys())
    leaf_nodes, remaining_node_indices = _remove_leaf_nodes(graph, node_to_id, node_indices)
    upload_order.extend(leaf_nodes)
    stash_counter = 0
    while remaining_node_indices:
        cycle = list(rx.digraph_find_cycle(graph))
        links_to_remove = _find_cheapest_outgoing_links(graph, cycle, edges)
        stash_counter += len(links_to_remove)
        _remove_edges_to_stash(
            graph=graph,
            edges_to_remove=links_to_remove,
            all_edges=edges,
            remaining_nodes=remaining_node_indices,
        )
        stash_lookup = _add_stash_to_lookup_dict(stash_lookup, [x.link_object for x in links_to_remove])
        leaf_nodes, remaining_node_indices = _remove_leaf_nodes(graph, node_to_id, remaining_node_indices)
        upload_order.extend(leaf_nodes)
    return stash_lookup, upload_order, stash_counter



from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import cast

import regex
from lxml import etree
from regex import Pattern

from dsp_tools.commands.xmlupload.models.ontology_lookup_models import AllowedEncodings
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import ProjectOntosInformation
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import PropertyTextValueTypes
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import TextValueData
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import get_text_value_types_of_properties_from_onto
from dsp_tools.commands.xmlupload.models.ontology_lookup_models import make_project_onto_information
from dsp_tools.commands.xmlupload.models.ontology_problem_models import InvalidOntologyElementsInData
from dsp_tools.commands.xmlupload.models.ontology_problem_models import InvalidTextValueEncodings
from dsp_tools.commands.xmlupload.ontology_client import OntologyClient
from dsp_tools.models.exceptions import InputError

defaultOntologyColon: Pattern[str] = regex.compile(r"^:\w+$")
knoraUndeclared: Pattern[str] = regex.compile(r"^\w+$")
genericPrefixedOntology: Pattern[str] = regex.compile(r"^[\w\-]+:\w+$")
KNORA_BASE_PROPERTIES = {
    "bitstream",
    "iiif-uri",
    "isSegmentOf",
    "hasSegmentBounds",
    "hasTitle",
    "hasComment",
    "hasDescription",
    "hasKeyword",
    "relatesTo",
}


def do_xml_consistency_check_with_ontology(onto_client: OntologyClient, root: etree._Element) -> None:
    """
    This function takes an OntologyClient and the root of an XML.
    It retrieves the ontologies from the server.
    It analyses if any classes or properties are used that are not in the ontology.
    It analyses if any properties have encodings that are not consistent with those specified in the ontology.

     Args:
         onto_client: client for the ontology retrieval
         root: root of the XML

     Raises:
         InputError: if there are any invalid properties or classes and/or text values with wrong encodings.
    """
    cls_prop_lookup, text_value_encoding_lookup = _get_onto_lookups(onto_client)
    classes_in_data, properties_in_data = _get_all_classes_and_properties_from_data(root)
    problem_str = ""
    problem_str += _check_all_classes_and_properties_in_onto(classes_in_data, properties_in_data, cls_prop_lookup)
    problem_str += _check_correctness_all_text_value_encodings(root, text_value_encoding_lookup)
    if problem_str:
        raise InputError(problem_str)


def _get_onto_lookups(onto_client: OntologyClient) -> tuple[ProjectOntosInformation, PropertyTextValueTypes]:
    ontos = onto_client.get_all_project_ontologies_from_server()
    text_value_encoding_lookup = get_text_value_types_of_properties_from_onto(ontos, onto_client.default_ontology)
    ontos["knora-api"] = onto_client.get_knora_api_ontology_from_server()
    return make_project_onto_information(onto_client.default_ontology, ontos), text_value_encoding_lookup


def _check_all_classes_and_properties_in_onto(
    classes_in_data: dict[str, list[str]],
    properties_in_data: dict[str, list[str]],
    onto_check_info: ProjectOntosInformation,
) -> str:
    class_problems = _find_all_class_types_in_onto(classes_in_data, onto_check_info)
    property_problems = _find_all_properties_in_onto(properties_in_data, onto_check_info)
    if not class_problems and not property_problems:
        return ""
    problems = InvalidOntologyElementsInData(
        classes=class_problems, properties=property_problems, ontos_on_server=list(onto_check_info.onto_lookup.keys())
    )
    msg, df = problems.execute_problem_protocol()
    if df is not None:
        csv_file = f"XML_syntax_errors_{datetime.now().strftime('%Y-%m-%d_%H%M%S')}.csv"
        df.to_csv(path_or_buf=Path(Path.cwd(), csv_file), index=False)
        msg += (
            "\n\n---------------------------------------\n\n"
            f"\nAll the problems are listed in the file: '{Path.cwd()}/{csv_file}'\n"
        )
    return msg


def _get_all_classes_and_properties_from_data(
    root: etree._Element,
) -> tuple[dict[str, list[str]], dict[str, list[str]]]:
    cls_dict = _get_all_class_types_and_ids_from_data(root)
    prop_dict: defaultdict[str, list[str]] = defaultdict(list)
    for resource in root.iterchildren(tag="resource"):
        prop_dict = _get_all_property_names_and_resource_ids_one_resource(resource, prop_dict)
    return cls_dict, prop_dict


def _get_all_class_types_and_ids_from_data(root: etree._Element) -> dict[str, list[str]]:
    cls_dict: dict[str, list[str]] = {}
    for resource in root.iterchildren(tag="resource"):
        restype = resource.attrib["restype"]
        if restype in cls_dict:
            cls_dict[restype].append(resource.attrib["id"])
        else:
            cls_dict[restype] = [resource.attrib["id"]]
    return cls_dict


def _get_all_property_names_and_resource_ids_one_resource(
    resource: etree._Element, prop_dict: defaultdict[str, list[str]]
) -> defaultdict[str, list[str]]:
    for prop in resource.iterchildren():
        if prop.tag in KNORA_BASE_PROPERTIES:
            continue
        prop_name = prop.attrib["name"]
        prop_dict[prop_name].append(resource.attrib["id"])
    return prop_dict


def _find_all_class_types_in_onto(
    classes: dict[str, list[str]], onto_check_info: ProjectOntosInformation
) -> list[tuple[str, list[str], str]]:
    problem_list = []
    for cls_type, ids in classes.items():
        if problem := _find_one_class_type_in_onto(cls_type, onto_check_info):
            problem_list.append((cls_type, ids, problem))
    return problem_list


def _find_one_class_type_in_onto(cls_type: str, onto_check_info: ProjectOntosInformation) -> str | None:
    prefix, cls_ = _get_separate_prefix_and_iri_from_onto_prop_or_cls(cls_type, onto_check_info.default_ontology_prefix)
    if not prefix:
        return "Class name does not follow a known ontology pattern"
    if onto := onto_check_info.onto_lookup.get(prefix):
        return "Invalid Class Type" if cls_ not in onto.classes else None
    else:
        return "Unknown ontology prefix"


def _find_all_properties_in_onto(
    properties: dict[str, list[str]], onto_check_info: ProjectOntosInformation
) -> list[tuple[str, list[str], str]]:
    problem_list = []
    for prop_name, ids in properties.items():
        if problem := _find_one_property_in_onto(prop_name, onto_check_info):
            problem_list.append((prop_name, ids, problem))
    return problem_list


def _find_one_property_in_onto(prop_name: str, onto_check_info: ProjectOntosInformation) -> str | None:
    prefix, prop = _get_separate_prefix_and_iri_from_onto_prop_or_cls(
        prop_name, onto_check_info.default_ontology_prefix
    )
    if not prefix:
        return "Property name does not follow a known ontology pattern"
    if onto := onto_check_info.onto_lookup.get(prefix):
        return "Invalid Property" if prop not in onto.properties else None
    else:
        return "Unknown ontology prefix"


def _get_separate_prefix_and_iri_from_onto_prop_or_cls(
    prop_or_cls: str, default_ontology_prefix: str
) -> tuple[str, ...] | tuple[None, None]:
    if defaultOntologyColon.match(prop_or_cls):
        return default_ontology_prefix, prop_or_cls.lstrip(":")
    elif knoraUndeclared.match(prop_or_cls):
        return "knora-api", prop_or_cls
    elif genericPrefixedOntology.match(prop_or_cls):
        return tuple(prop_or_cls.split(":"))
    else:
        return None, None


def _check_correctness_all_text_value_encodings(root: etree._Element, text_prop_look_up: PropertyTextValueTypes) -> str:
    """
    This function analyses if all the encodings for the `<text>` elements are correct
    with respect to the specification in the ontology.

    For example, if the ontology specifies that `:hasSimpleText` is without mark-up,
    the encoding has to be `utf8`.


    This is correct:
    ```
    <text-prop name=":hasSimpleText">
        <text encoding="utf8">Dies ist ein einfacher Text ohne Markup</text>
    </text-prop>
    ```

    This is wrong:
    ```
    <text-prop name=":hasSimpleText">
        <text encoding="xml">Dies ist ein einfacher Text ohne Markup</text>
    </text-prop>
    ```

    The accepted encodings are `xml` or `utf8`

    Args:
        root: root of the data xml document
        text_prop_look_up: a lookup containing the property names and their specified types

    Returns:
        A string communicating the problem, if there are none the string is empty.
    """
    text_values_in_data = _get_all_ids_and_props_and_encodings_from_root(root)
    invalid_text_values = [x for x in text_values_in_data if not _check_correctness_of_one_prop(x, text_prop_look_up)]
    if not invalid_text_values:
        return ""
    msg, df = InvalidTextValueEncodings(invalid_text_values).execute_problem_protocol()
    if df is not None:
        csv_file = Path(f"text_value_encoding_errors_{datetime.now().strftime('%Y-%m-%d_%H%M%S')}.csv")
        df.to_csv(path_or_buf=csv_file, index=False)
        msg += (
            "\n\n---------------------------------------\n\n"
            f"All the problems are listed in the file: '{csv_file.absolute()}'"
        )
    return msg


def _get_all_ids_and_props_and_encodings_from_root(root: etree._Element) -> list[TextValueData]:
    res_list = []
    for res_input in root.iterchildren(tag="resource"):
        res_list.extend(_get_id_and_props_and_encodings_from_one_resource(res_input))
    return res_list


def _get_id_and_props_and_encodings_from_one_resource(resource: etree._Element) -> list[TextValueData]:
    res_id = resource.attrib["id"]
    res_type = resource.attrib["restype"]
    return [
        _get_prop_and_encoding_from_one_property(res_id, res_type, child)
        for child in resource.iterchildren(tag="text-prop")
    ]


def _get_prop_and_encoding_from_one_property(res_id: str, res_type: str, prop: etree._Element) -> TextValueData:
    prop_name = prop.attrib["name"]
    encoding = cast(AllowedEncodings, prop[0].attrib["encoding"])
    return TextValueData(res_id, res_type, prop_name, encoding)


def _check_correctness_of_one_prop(text_val: TextValueData, text_prop_look_up: PropertyTextValueTypes) -> bool:
    match text_val.encoding:
        case "xml":
            return text_val.property_name in text_prop_look_up.formatted_text_props
        case "utf8":
            return text_val.property_name in text_prop_look_up.unformatted_text_props
        case _:
            return False



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Protocol


@dataclass(frozen=True)
class SerialiseProperty:
    property_name: str
    values: list[SerialiseValue]

    def serialise(self) -> dict[str, Any]:
        """Serialise the property and all its values."""
        return {self.property_name: [value.serialise() for value in self.values]}


class SerialiseValue(Protocol):
    """A value to be serialised."""

    value: str
    permissions: str | None
    comment: str | None

    def serialise(self) -> dict[str, Any]:
        """Serialise the value."""


@dataclass(frozen=True)
class SerialiseURI(SerialiseValue):
    """A URI to be serialised."""

    value: str
    permissions: str | None
    comment: str | None

    def serialise(self) -> dict[str, Any]:
        """Serialise the URI value."""

        serialised = {
            "@type": "knora-api:UriValue",
            "knora-api:uriValueAsUri": {
                "@type": "xsd:anyURI",
                "@value": self.value,
            },
        }
        if self.comment:
            serialised["knora-api:valueHasComment"] = self.comment
        if self.permissions:
            serialised["knora-api:hasPermissions"] = self.permissions
        return serialised



from unittest.mock import Mock
from unittest.mock import patch

import pytest
from requests import RequestException
from requests import Response

from dsp_tools.commands.xmlupload.iiif_uri_validator import IIIFUriValidator
from dsp_tools.commands.xmlupload.models.input_problems import IIIFUriProblem


@pytest.fixture
def response_404() -> Response:
    mock_response = Response()
    mock_response.status_code = 404
    mock_response._content = b"This is the response text."
    return mock_response


@pytest.fixture
def response_200() -> Response:
    mock_response = Response()
    mock_response.status_code = 200
    mock_response._content = b"This is the response text."
    return mock_response


@pytest.fixture
def request_exception() -> RequestException:
    return RequestException("This is the request exception.")


@pytest.fixture
def empty_validator() -> IIIFUriValidator:
    return IIIFUriValidator([])


@pytest.mark.parametrize(
    ("uri", "expected"),
    [
        (
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2",
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/info.json",
        ),
        (
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/2048,0,1298,2048/649,1024/0/default.jpg",
            "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/info.json",
        ),
        (
            "https://iiif.io/api/image/3.0/example/reference/1-newspaper-p2/full/max/0/default.webp",
            "https://iiif.io/api/image/3.0/example/reference/1-newspaper-p2/info.json",
        ),
        (
            "https://newspapers.library.wales/iiif/2.0/image/4497470/512,2048,512,512/256,/0/default.jpg",
            "https://newspapers.library.wales/iiif/2.0/image/4497470/info.json",
        ),
        (
            "https://example.org/image-service/abcd1234/full/max/0/default.jpg",
            "https://example.org/image-service/abcd1234/info.json",
        ),
        (
            "https://www.example.org/prefix1/abcd1234/80,15,60,75/full/0/native",
            "https://www.example.org/prefix1/abcd1234/info.json",
        ),
        (
            "http://www.example.org/prefix1/abcd1234/80,15,60.6,75/full/0/native.jpg",
            "http://www.example.org/prefix1/abcd1234/info.json",
        ),
        ("bla", "bla/info.json"),
        ("bla/", "bla/info.json"),
    ],
)
def test_make_info_json_uri_correct(empty_validator: IIIFUriValidator, uri: str, expected: str) -> None:
    assert empty_validator._make_info_json_uri(uri) == expected


@patch("dsp_tools.commands.xmlupload.iiif_uri_validator.IIIFUriValidator._make_network_call")
def test_validate_with_exception(
    mock_network_call: Mock, request_exception: RequestException, empty_validator: IIIFUriValidator
) -> None:
    mock_network_call.return_value = request_exception
    result = empty_validator._validate_one_uri("http://example.com")
    assert isinstance(result, IIIFUriProblem)
    assert result.uri == "http://example.com"
    assert not result.regex_has_passed
    assert not result.status_code
    assert result.original_text == "This is the request exception."
    assert result.raised_exception_name == "RequestException"


@patch("dsp_tools.commands.xmlupload.iiif_uri_validator.IIIFUriValidator._make_network_call")
def test_validate_with_bad_status_code(
    mock_network_call: Mock, response_404: Response, empty_validator: IIIFUriValidator
) -> None:
    mock_network_call.return_value = response_404
    result = empty_validator._validate_one_uri("http://example.com")
    assert isinstance(result, IIIFUriProblem)
    assert result.uri == "http://example.com"
    assert not result.regex_has_passed
    assert result.status_code == 404
    assert result.original_text == "This is the response text."
    assert not result.raised_exception_name


@patch("dsp_tools.commands.xmlupload.iiif_uri_validator.IIIFUriValidator._make_network_call")
def test_validate_with_good_status_code(
    mock_network_call: Mock, response_200: Response, empty_validator: IIIFUriValidator
) -> None:
    mock_network_call.return_value = response_200
    result = empty_validator._validate_one_uri(
        "https://iiif.dasch.swiss/0811/1Oi7mdiLsG7-FmFgp0xz2xU.jp2/full/837,530/0/default.jp2"
    )
    assert not result


@patch("dsp_tools.commands.xmlupload.iiif_uri_validator.IIIFUriValidator._make_network_call")
def test_validate_with_failed_regex_good_status_code(
    mock_network_call: Mock, response_200: Response, empty_validator: IIIFUriValidator
) -> None:
    mock_network_call.return_value = response_200
    result = empty_validator._validate_one_uri("http://example.com")
    assert isinstance(result, IIIFUriProblem)
    assert result.uri == "http://example.com"
    assert not result.regex_has_passed
    assert result.status_code == 200
    assert result.original_text == "This is the response text."
    assert not result.raised_exception_name


if __name__ == "__main__":
    pytest.main([__file__])



from __future__ import annotations

import warnings
from abc import ABC
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path

import pandas as pd
from rdflib import Graph

from dsp_tools.commands.validate_data.models.validation import UnexpectedComponent
from dsp_tools.models.custom_warnings import DspToolsUserWarning

LIST_SEPARATOR = "\n    - "
GRAND_SEPARATOR = "\n\n----------------------------\n"


@dataclass
class UnexpectedResults:
    components: list[UnexpectedComponent]

    def save_inform_user(self, results_graph: Graph, shacl: Graph, data: Graph) -> None:
        cwdr = Path.cwd()
        prefix = f"{datetime.now()!s}_"
        unique_components = list(set(x.component_type for x in self.components))
        components = sorted(unique_components)
        msg = (
            f"Unexpected violations were found in the validation results:"
            f"{LIST_SEPARATOR}{LIST_SEPARATOR.join(components)}\n"
            f"Please contact the development team with the files starting with the timestamp '{prefix}' "
            f"in the directory '{cwdr}'."
        )
        save_path = cwdr / f"{prefix}validation_result.ttl"
        results_graph.serialize(save_path)
        shacl_p = cwdr / f"{prefix}shacl.ttl"
        shacl.serialize(shacl_p)
        data_p = cwdr / f"{prefix}data.ttl"
        data.serialize(data_p)
        warnings.warn(DspToolsUserWarning(msg))


@dataclass
class AllProblems:
    problems: list[InputProblem]
    unexpected_results: UnexpectedResults | None

    def get_msg(self, file_path: Path) -> str:
        coll = self._make_collection()
        title_msg = f"\nDuring the validation of the data {len(self.problems)} errors were found:\n\n"
        if len(self.problems) > 50:
            out_path = file_path.parent / f"{file_path.stem}_validation_errors.csv"
            self._save_as_csv(out_path)
            out_message = (
                title_msg + f"Due to the large number or errors, the validation errors were saved at:\n{out_path}"
            )
        else:
            msg = [x.get_msg() for x in coll]
            out_message = title_msg + GRAND_SEPARATOR.join(msg)
        return out_message

    def _make_collection(self) -> list[ResourceProblemCollection]:
        d = defaultdict(list)
        for e in self.problems:
            d[e.res_id].append(e)
        collection_list = []
        for k, v in d.items():
            collection_list.append(ResourceProblemCollection(k, v))
        return sorted(collection_list, key=lambda x: x.res_id)

    def _save_as_csv(self, out_path: Path) -> None:
        all_problems = [x.to_dict() for x in self.problems]
        df = pd.DataFrame.from_records(all_problems)
        df = df.sort_values(by=["Resource Type", "Resource ID", "Property"])
        df.to_csv(out_path, index=False)


@dataclass
class ResourceProblemCollection:
    res_id: str
    problems: list[InputProblem]

    def get_msg(self) -> str:
        prop_msg = self._msg_for_properties()
        return f"Resource ID: {self.res_id} | Resource Type: {self.problems[0].res_type}\n{prop_msg}"

    def _msg_for_properties(self) -> str:
        grouped = self._make_collection()
        out_list = []
        for prop_name, problem in grouped.items():
            problem_list = [x.get_msg() for x in problem]
            out_list.append((prop_name, f"{prop_name}{LIST_SEPARATOR}{LIST_SEPARATOR.join(problem_list)}"))
        sorted_list = sorted(out_list, key=lambda x: x[0])
        return "\n".join([x[1] for x in sorted_list])

    def _make_collection(self) -> dict[str, list[InputProblem]]:
        grouped_dict = defaultdict(list)
        for problem in self.problems:
            grouped_dict[problem.sort_value()].append(problem)
        return grouped_dict

    def sort_value(self) -> str:
        return self.res_id


@dataclass
class InputProblem(ABC):
    res_id: str
    res_type: str
    prop_name: str

    @property
    def problem(self) -> str:
        raise NotImplementedError

    def get_msg(self) -> str:
        raise NotImplementedError

    def to_dict(self) -> dict[str, str]:
        raise NotImplementedError

    def _base_dict(self) -> dict[str, str]:
        return {
            "Resource Type": self.res_type,
            "Resource ID": self.res_id,
            "Property": self.prop_name,
            "Problem": self.problem,
        }

    def sort_value(self) -> str:
        raise NotImplementedError


#######################
# Cardinality Violation


@dataclass
class MaxCardinalityViolation(InputProblem):
    expected_cardinality: str

    @property
    def problem(self) -> str:
        return "Maximum Cardinality Violation"

    def get_msg(self) -> str:
        return f"{self.problem} | Expected Cardinality: {self.expected_cardinality}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = f"Cardinality: {self.expected_cardinality}"
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class MinCardinalityViolation(InputProblem):
    expected_cardinality: str

    @property
    def problem(self) -> str:
        return "Minimum Cardinality Violation"

    def get_msg(self) -> str:
        return f"{self.problem} | Expected Cardinality: {self.expected_cardinality}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = f"Cardinality: {self.expected_cardinality}"
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


@dataclass
class NonExistentCardinalityViolation(InputProblem):
    @property
    def problem(self) -> str:
        return "The resource class does not have a cardinality for this property."

    def get_msg(self) -> str:
        return self.problem

    def to_dict(self) -> dict[str, str]:
        return self._base_dict()

    def sort_value(self) -> str:
        return self.prop_name


#######################
# Value Type Violation


@dataclass
class ValueTypeViolation(InputProblem):
    actual_type: str
    expected_type: str

    @property
    def problem(self) -> str:
        return "Value Type Mismatch"

    def get_msg(self) -> str:
        return f"{self.problem}, Actual Type: {self.actual_type} | Expected Type: {self.expected_type}"

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected_type
        problm_dict["Actual"] = self.actual_type
        return problm_dict

    def sort_value(self) -> str:
        return self.prop_name


#######################
# Content Violation


@dataclass
class ContentRegexViolation(InputProblem):
    expected_format: str
    actual_content: str | None

    @property
    def problem(self) -> str:
        return "Wrong Content Format"

    def get_msg(self) -> str:
        msg = f"{self.problem}, Expected Format: {self.expected_format}"
        if self.actual_content:
            msg += f" | Actual Content: '{self._short_content()}'"
        return msg

    def to_dict(self) -> dict[str, str]:
        problm_dict = self._base_dict()
        problm_dict["Expected"] = self.expected_format
        if self.actual_content:
            problm_dict["Actual"] = self._short_content()
        return problm_dict

    def _short_content(self) -> str:
        if not self.actual_content:
            return ""
        if len(self.actual_content) > 15:
            return f"{self.actual_content[:15]}[...]"
        return self.actual_content

    def sort_value(self) -> str:
        return self.prop_name



from dataclasses import dataclass
from dataclasses import field


@dataclass
class ListNodeNames:
    """
    Contains the information to construct a list

    Args:
        lists_with_previous_cell_values: used to ensure that there are no duplicate node names
        previous_node_names: used to ensure that there are no duplicate node names
    """

    lists_with_previous_cell_values: list[list[str]] = field(default_factory=list)
    previous_node_names: list[str] = field(default_factory=list)



from enum import Enum
from enum import auto


class Permissions(Enum):
    DOAP = ""
    OPEN = "open"
    RESTRICTED = "restricted"
    RESTRICTED_VIEW = "restricted-view"


class NewlineReplacement(Enum):
    NONE = auto()
    PARAGRAPH = auto()
    LINEBREAK = auto()



from pathlib import Path

import pytest

from dsp_tools.commands import rosetta


def test_update_repo_before_cloning(tmp_path: Path) -> None:
    rosetta_folder = tmp_path / "082E-rosetta-scripts"
    is_rosetta_up_to_date = rosetta._update_possibly_existing_repo(rosetta_folder=rosetta_folder)
    assert not is_rosetta_up_to_date


def test_update_repo_after_cloning(tmp_path: Path) -> None:
    rosetta_folder = tmp_path / "082E-rosetta-scripts"
    rosetta._clone_repo(rosetta_folder=rosetta_folder, enclosing_folder=tmp_path)
    is_rosetta_up_to_date = rosetta._update_possibly_existing_repo(rosetta_folder=rosetta_folder)
    assert is_rosetta_up_to_date


if __name__ == "__main__":
    pytest.main([__file__])



from dataclasses import dataclass

import requests

from dsp_tools.commands.xmlupload.models.input_problems import AllIIIFUriProblems
from dsp_tools.commands.xmlupload.models.input_problems import IIIFUriProblem
from dsp_tools.utils.uri_util import is_iiif_uri


@dataclass(frozen=True)
class IIIFUriValidator:
    """Client handling communication with external IIIF-servers to do a health check."""

    uri_list: list[str]

    def validate(self) -> AllIIIFUriProblems | None:
        """Validate the URI and return a list of problems if any."""
        iiif_uri_problems = [res for uri in self.uri_list if (res := self._validate_one_uri(uri)) is not None]
        if iiif_uri_problems:
            return AllIIIFUriProblems(problems=iiif_uri_problems)
        return None

    def _validate_one_uri(self, uri: str) -> IIIFUriProblem | None:
        """Check if the IIIF-server is reachable. If not, it returns information for error message."""
        regex_has_passed = is_iiif_uri(uri)
        response = self._make_network_call(uri)
        if isinstance(response, Exception):
            return IIIFUriProblem(
                uri=uri,
                regex_has_passed=regex_has_passed,
                raised_exception_name=response.__class__.__name__,
                original_text=str(response),
            )
        match response.ok, regex_has_passed:
            case True, True:
                return None
            case _:
                return IIIFUriProblem(
                    uri=uri,
                    regex_has_passed=regex_has_passed,
                    status_code=response.status_code,
                    original_text=response.text,
                )

    def _make_network_call(self, uri: str) -> requests.Response | Exception:
        info_json_uri = self._make_info_json_uri(uri)
        try:
            return requests.get(
                url=info_json_uri,
                headers={"Content-Type": "application/ld+json"},
                timeout=10,
            )
        except Exception as e:  # noqa: BLE001 (blind-except)
            return e

    def _make_info_json_uri(self, uri: str) -> str:
        splt = uri.split("/")
        if len(splt) < 5:
            info_uri = uri.rstrip("/")
        else:
            info_uri = "/".join(splt[:-4])
        return f"{info_uri}/info.json"



import warnings
from copy import deepcopy
from pathlib import Path

from lxml import etree
from rdflib import SH
from rdflib import Graph
from termcolor import cprint

from dsp_tools.commands.validate_data.api_connection import OntologyConnection
from dsp_tools.commands.validate_data.api_connection import ShaclValidator
from dsp_tools.commands.validate_data.deserialise_input import deserialise_xml
from dsp_tools.commands.validate_data.make_data_rdf import make_data_rdf
from dsp_tools.commands.validate_data.models.data_deserialised import ProjectDeserialised
from dsp_tools.commands.validate_data.models.data_rdf import DataRDF
from dsp_tools.commands.validate_data.models.validation import RDFGraphs
from dsp_tools.commands.validate_data.models.validation import ValidationReports
from dsp_tools.commands.validate_data.reformat_validaton_result import reformat_validation_graph
from dsp_tools.commands.validate_data.sparql.cardinality_shacl import construct_cardinality_node_shapes
from dsp_tools.commands.validate_data.sparql.content_shacl import construct_content_shapes_graph
from dsp_tools.models.custom_warnings import DspToolsUserWarning
from dsp_tools.utils.xml_utils import parse_xml_file
from dsp_tools.utils.xml_utils import remove_comments_from_element_tree
from dsp_tools.utils.xml_utils import transform_into_localnames
from dsp_tools.utils.xml_validation import validate_xml

LIST_SEPARATOR = "\n    - "


def validate_data(filepath: Path, api_url: str, dev_route: bool, save_graphs: bool) -> bool:  # noqa: ARG001 (unused argument)
    """
    Takes a file and project information and validates it against the ontologies on the server.

    Args:
        filepath: path to the xml data file
        api_url: url of the api host
        dev_route: if this flag is set features that are still in development will be used
        save_graphs: if this flag is set, all the graphs will be saved in a folder

    Returns:
        true unless it crashed
    """
    _inform_about_experimental_feature()
    data_rdf, shortcode = _get_data_info_from_file(filepath, api_url)
    onto_con = OntologyConnection(api_url, shortcode)
    rdf_graphs = _create_graphs(onto_con, data_rdf)
    generic_filepath = Path()
    if save_graphs:
        generic_filepath = _save_graphs(filepath, rdf_graphs)
    val = ShaclValidator(api_url)
    report = _validate(val, rdf_graphs)
    if save_graphs:
        if report.content_validation:
            report.content_validation.serialize(f"{generic_filepath}_VALIDATION_REPORT_CONTENT.ttl")
        if report.cardinality_validation:
            report.cardinality_validation.serialize(f"{generic_filepath}_VALIDATION_REPORT_CARD.ttl")
    if report.conforms:
        cprint("\n   Validation passed!   ", color="green", attrs=["bold", "reverse"])
    else:
        reformatted = reformat_validation_graph(report)
        problem_msg = reformatted.get_msg(filepath)
        cprint("\n   Validation errors found!   ", color="light_red", attrs=["bold", "reverse"])
        print(problem_msg)
        if reformatted.unexpected_results:
            report_graphs = Graph()
            if report.content_validation:
                report_graphs += report.content_validation
            if report.cardinality_validation:
                report_graphs += report.cardinality_validation
            reformatted.unexpected_results.save_inform_user(
                results_graph=report_graphs,
                shacl=report.shacl_graphs,
                data=rdf_graphs.data,
            )
    return True


def _inform_about_experimental_feature() -> None:
    what_is_validated = [
        "This is an experimental feature, it will change and be extended continuously. "
        "The following information of your data is being validated:",
        "Cardinalities",
    ]
    warnings.warn(DspToolsUserWarning(LIST_SEPARATOR.join(what_is_validated)))


def _create_graphs(onto_con: OntologyConnection, data_rdf: DataRDF) -> RDFGraphs:
    ontologies = _get_project_ontos(onto_con)
    knora_ttl = onto_con.get_knora_api()
    kag = Graph()
    kag.parse(data=knora_ttl, format="ttl")
    onto_for_construction = deepcopy(ontologies) + kag
    card_shapes = construct_cardinality_node_shapes(onto_for_construction)
    content_shapes = construct_content_shapes_graph(onto_for_construction)
    api_shapes = Graph()
    api_shapes.parse("src/dsp_tools/resources/validate_data/api-shapes.ttl")
    content_shapes += api_shapes
    return RDFGraphs(
        data=data_rdf.make_graph(), ontos=ontologies, cardinality_shapes=card_shapes, content_shapes=content_shapes
    )


def _get_project_ontos(onto_con: OntologyConnection) -> Graph:
    all_ontos = onto_con.get_ontologies()
    onto_g = Graph()
    for onto in all_ontos:
        og = Graph()
        og.parse(data=onto, format="ttl")
        onto_g += og
    return onto_g


def _save_graphs(filepath: Path, rdf_graphs: RDFGraphs) -> Path:
    parent_directory = filepath.parent
    new_directory = parent_directory / "graphs"
    new_directory.mkdir(exist_ok=True)
    cprint(f"\n   Saving graphs to {new_directory}   ", color="light_blue", attrs=["bold", "reverse"])
    generic_filepath = new_directory / filepath.stem
    rdf_graphs.ontos.serialize(f"{generic_filepath}_ONTO.ttl")
    rdf_graphs.cardinality_shapes.serialize(f"{generic_filepath}_SHACL_CARD.ttl")
    rdf_graphs.content_shapes.serialize(f"{generic_filepath}_SHACL_CONTENT.ttl")
    rdf_graphs.data.serialize(f"{generic_filepath}_DATA.ttl")
    onto_data = rdf_graphs.data + rdf_graphs.ontos
    onto_data.serialize(f"{generic_filepath}_ONTO_DATA.ttl")
    return generic_filepath


def _validate(validator: ShaclValidator, rdf_graphs: RDFGraphs) -> ValidationReports:
    card_shacl = rdf_graphs.get_cardinality_shacl_str()
    card_data = rdf_graphs.get_cardinality_data_str()
    card_results = validator.validate(card_data, card_shacl)
    card_conforms = bool(next(card_results.objects(None, SH.conforms)))

    content_shacl = rdf_graphs.get_content_shacl_str()
    content_data = rdf_graphs.get_content_data_str()
    content_results = validator.validate(content_data, content_shacl)
    content_conforms = bool(next(content_results.objects(None, SH.conforms)))

    conforms = all([content_conforms, card_conforms])
    card_report: Graph | None = None
    content_report: Graph | None = None
    if not card_conforms:
        card_report = card_results
    if not content_conforms:
        content_report = content_results
    return ValidationReports(
        conforms=conforms,
        content_validation=content_report,
        cardinality_validation=card_report,
        shacl_graphs=rdf_graphs.cardinality_shapes + rdf_graphs.content_shapes,
        data_graph=rdf_graphs.data,
    )


def _get_data_info_from_file(file: Path, api_url: str) -> tuple[DataRDF, str]:
    cleaned_root = _parse_and_clean_file(file, api_url)
    deserialised: ProjectDeserialised = deserialise_xml(cleaned_root)
    rdf_data: DataRDF = make_data_rdf(deserialised.data)
    return rdf_data, deserialised.info.shortcode


def _parse_and_clean_file(file: Path, api_url: str) -> etree._Element:
    root = parse_xml_file(file)
    root = remove_comments_from_element_tree(root)
    validate_xml(root)
    root = transform_into_localnames(root)
    return _replace_namespaces(root, api_url)


def _replace_namespaces(root: etree._Element, api_url: str) -> etree._Element:
    with open("src/dsp_tools/resources/validate_data/replace_namespace.xslt", "rb") as xslt_file:
        xslt_data = xslt_file.read()
    shortcode = root.attrib["shortcode"]
    default_ontology = root.attrib["default-ontology"]
    namespace = f"{api_url}/ontology/{shortcode}/{default_ontology}/v2#"
    xslt_root = etree.XML(xslt_data)
    transform = etree.XSLT(xslt_root)
    replacement_value = etree.XSLT.strparam(namespace)
    return transform(root, replacementValue=replacement_value).getroot()



from typing import Any

import numpy as np
import pandas as pd
import pytest

from dsp_tools.xmllib.value_checkers import find_geometry_problem
from dsp_tools.xmllib.value_checkers import is_bool_like
from dsp_tools.xmllib.value_checkers import is_color
from dsp_tools.xmllib.value_checkers import is_date
from dsp_tools.xmllib.value_checkers import is_decimal
from dsp_tools.xmllib.value_checkers import is_dsp_ark
from dsp_tools.xmllib.value_checkers import is_dsp_iri
from dsp_tools.xmllib.value_checkers import is_geoname
from dsp_tools.xmllib.value_checkers import is_integer
from dsp_tools.xmllib.value_checkers import is_string_like
from dsp_tools.xmllib.value_checkers import is_timestamp


@pytest.mark.parametrize("val", ["false", "0", "0.0", "no", "true", "1", "1.0", "yes", False, True])
def test_is_bool_correct(val: Any) -> None:
    assert is_bool_like(val)


@pytest.mark.parametrize("val", ["none", "asdfioh", None, pd.NA])
def test_is_bool_wrong(val: Any) -> None:
    assert not is_bool_like(val)


@pytest.mark.parametrize("val", ["#234567", "#abcdef", "#123abc"])
def test_is_color_correct(val: str) -> None:
    assert is_color(val)


@pytest.mark.parametrize("val", [12334, "123abc", "234567#"])
def test_is_color_wrong(val: Any) -> None:
    assert not is_color(val)


@pytest.mark.parametrize(
    "val",
    [
        "GREGORIAN:CE:0476-09-04:CE:0476-09-04",
        "GREGORIAN:BC:2000:BC:1000",
        "JULIAN:CE:0476-09-04:CE:0476-09-04",
        "JULIAN:BC:2000:BC:1000",
        "ISLAMIC:CE:0476-09-04:CE:0476-09-04",
        "ISLAMIC:BC:2000:BC:1000",
        "GREGORIAN:0476-09-04:0476-09-04",
        "CE:0476-09-04:CE:0476-09-04",
    ],
)
def test_is_date_correct(val: str) -> None:
    assert is_date(val)


@pytest.mark.parametrize(
    "val",
    [
        "OTHER:BC:2000:BC:1000",
        "GREGORIAN:other:2000:other:1000",
        "GREGORIAN::CE:0476-09-04:CE:0476-09-04",
        "GREGORIAN:CE:0476-09-04:CE::0476-09-04",
        "GREGORIAN:CE:0476-09-04::CE:0476-09-04",
        "GREGORIAN:CE::0476-09-04:CE:0476-09-04",
    ],
)
def test_is_date_wrong(val: Any) -> None:
    assert not is_date(val)


@pytest.mark.parametrize("val", ["00001", 23412])
def test_is_geoname_correct(val: Any) -> None:
    assert is_geoname(val)


@pytest.mark.parametrize("val", [122.2, "asdf"])
def test_is_geoname_wrong(val: Any) -> None:
    assert not is_geoname(val)


@pytest.mark.parametrize("val", [1.2, "1.432", 1, "1", -1.1, "1e-1", "1e2"])
def test_is_decimal_correct(val: Any) -> None:
    assert is_decimal(val)


@pytest.mark.parametrize("val", ["asf", np.nan, pd.NA, True, False])
def test_is_decimal_wrong(val: Any) -> None:
    assert not is_decimal(val)


@pytest.mark.parametrize("val", [1, "1", -1])
def test_is_integer_correct(val: Any) -> None:
    assert is_integer(val)


@pytest.mark.parametrize("val", [1.2, "1.2", "wdasd", True, False, "1e2"])
def test_is_integer_wrong(val: Any) -> None:
    assert not is_integer(val)


@pytest.mark.parametrize("val", ["a", "None", "-", "1", "\n1", "עִבְרִית", "اَلْعَرَبِيَّةُ", "hello world"])
def test_is_string_correct(val: Any) -> None:
    assert is_string_like(val)


@pytest.mark.parametrize("val", [None, pd.NA, "", " ", "\t \n"])
def test_is_string_wrong(val: Any) -> None:
    assert not is_string_like(val)


@pytest.mark.parametrize(
    "val",
    [
        "2019-10-23T13:45:12.01-14:00",
        "2019-10-23T13:45:12-14:00",
        "2019-10-23T13:45:12Z",
        "2019-10-23T13:45:12-13:30",
        "2019-10-23T13:45:12+01:00",
        "2019-10-23T13:45:12.1111111+01:00",
        "2019-10-23T13:45:12.123456789012Z",
    ],
)
def test_is_timestamp_correct(val: Any) -> None:
    assert is_timestamp(val)


@pytest.mark.parametrize(
    "val",
    [
        True,
        10.0,
        5,
        "2019-10-2",
        "CE:1849:CE:1850",
        "2019-10-23T13:45:12.1234567890123Z",
        "2022",
        "GREGORIAN:CE:2014-01-31",
    ],
)
def test_is_timestamp_wrong(val: Any) -> None:
    assert not is_timestamp(val)


@pytest.mark.parametrize(
    "val",
    [
        '{"type": "rectangle", "lineColor": "#ff3333", "lineWidth": 2, '
        '"points": [{"x": 0.08, "y": 0.16}, {"x": 0.73, "y": 0.72}], "original_index": 0}',
        '{"type": "rectangle", "lineColor": "#000000", "lineWidth": 1, '
        '"points": [{"x": 0.10, "y": 0.10}, {"x": 0.10, "y": 0.10}], "original_index": 1}',
    ],
)
def test_is_geometry_correct(val: Any) -> None:
    assert not find_geometry_problem(val)


@pytest.mark.parametrize("val", ["100", 100, [0], '{"type": "polygon"}'])
def test_is_geometry_wrong(val: Any) -> None:
    assert find_geometry_problem(val)


@pytest.mark.parametrize("val", ["http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA"])
def test_is_dsp_iri_correct(val: Any) -> None:
    assert is_dsp_iri(val)


@pytest.mark.parametrize("val", ["http://www.example.org/prefix1/", "ark:/72163/4123-43xc6ivb931-a.2022829"])
def test_is_dsp_iri_wrong(val: Any) -> None:
    assert not is_dsp_iri(val)


@pytest.mark.parametrize("val", ["ark:/72163/4123-43xc6ivb931-a.2022829"])
def test_is_dsp_ark_correct(val: Any) -> None:
    assert is_dsp_ark(val)


@pytest.mark.parametrize("val", ["http://www.example.org/prefix1/", "http://rdfh.ch/4123/DiAmYQzQSzC7cdTo6OJMYA"])
def test_is_dsp_ark_wrong(val: Any) -> None:
    assert not is_dsp_ark(val)


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from rdflib import RDF
from rdflib import SH
from rdflib import Graph
from rdflib import Namespace

from dsp_tools.commands.validate_data.sparql.value_shacl import _add_property_shapes_to_class_shapes
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_link_value_shape
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_one_property_type_shape_based_on_object_type
from dsp_tools.commands.validate_data.sparql.value_shacl import _construct_one_property_type_text_value

ONTO = Namespace("http://0.0.0.0:3333/ontology/9999/onto/v2#")
API_SHAPES = Namespace("http://api.knora.org/ontology/knora-api/shapes/v2#")


def test_construct_one_property_type_shape_based_on_object_type(one_res_one_prop: Graph) -> None:
    res = _construct_one_property_type_shape_based_on_object_type(
        one_res_one_prop, "knora-api:BooleanValue", "api-shapes:BooleanValue_ClassShape"
    )
    assert len(res) == 3
    assert next(res.objects(ONTO.testBoolean_PropShape, SH.path)) == ONTO.testBoolean
    assert next(res.objects(ONTO.testBoolean_PropShape, RDF.type)) == SH.PropertyShape
    assert next(res.objects(ONTO.testBoolean_PropShape, SH.node)) == API_SHAPES.BooleanValue_ClassShape


def test_construct_link_value_shape(link_prop: Graph) -> None:
    res = _construct_link_value_shape(link_prop)
    assert len(res) == 3
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, SH.path)) == ONTO.testHasLinkTo
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, RDF.type)) == SH.PropertyShape
    assert next(res.objects(ONTO.testHasLinkTo_PropShape, SH.node)) == API_SHAPES.LinkValue_ClassShape


def test_construct_one_property_type_text_value(one_richtext_prop: Graph) -> None:
    res = _construct_one_property_type_text_value(
        one_richtext_prop, "salsah-gui:Richtext", "api-shapes:FormattedTextValue_ClassShape"
    )
    assert len(res) == 3
    assert next(res.objects(ONTO.testRichtext_PropShape, SH.path)) == ONTO.testRichtext
    assert next(res.objects(ONTO.testRichtext_PropShape, RDF.type)) == SH.PropertyShape
    assert next(res.objects(ONTO.testRichtext_PropShape, SH.node)) == API_SHAPES.FormattedTextValue_ClassShape


def test_add_property_shapes_to_class_shapes(card_1: Graph) -> None:
    res = _add_property_shapes_to_class_shapes(card_1)
    assert len(res) == 2
    assert next(res.objects(ONTO.ClassMixedCard_Shape, SH.property)) == ONTO.testBoolean_PropShape
    assert next(res.objects(ONTO.ClassMixedCard_Shape, SH.targetClass)) == ONTO.ClassMixedCard


if __name__ == "__main__":
    pytest.main([__file__])



from pathlib import Path

import pytest
from lxml import etree

from dsp_tools.commands.validate_data.validate_data import _parse_and_clean_file


@pytest.fixture
def data_xml() -> etree._Element:
    return _parse_and_clean_file(Path("testdata/validate-data/data/minimal_correct.xml"), "http://0.0.0.0:3333")
